/*
 *	Copyright  1994 by S.R.
 *
 *	Created: 23 Jan 1994
 *	Modified:	23 May 1994	14:28:11
 *
 * Make>> smake
 */

#include "/Debug.h"

#include "/File-HandlerClass.h"
#include "File-Handler.h"
#include "Handler.h"

#include "proto/MTools.h"
#include "proto/AsyncWork.h"
#include "proto/Misc.h"

ULONG DoFHM_ReplyPacket( Class *Cl, Object *O, struct TagItem TagP[], struct FileHandlerData * const Data)
{
	struct DosPacket *Packet;

	Packet = (struct DosPacket *)GetTagData( FHP_Packet, NULL, TagP);

	if ( Packet)
		ReplyPacket( Packet, Data->HandlerPort);
	else
		DBG_Error( KPrintF("%s, %s(), %l04d>E:Packet:%l08x\n", __FILE__, __FUNC__, __LINE__, Packet););

	return TRUE;
}


ULONG DoFHM_MakeVolumeNode( Class *Cl, Object *O, struct TagItem TagP[], struct FileHandlerData * const Data)
{
	struct DateStamp CurrentDate;
	struct DateStamp *VolumeDate;
	STRPTR VolumeName;

	DateStamp( &CurrentDate);
	VolumeName = (STRPTR)GetTagData( FHP_StrPtr, NULL, TagP);
	VolumeDate = (struct DateStamp *)GetTagData( FHP_DateStamp, (ULONG)&CurrentDate, TagP);
	if ( VolumeName)
		{
		struct DeviceList *VolumeNode;

		if ( VolumeNode = AllocVec( sizeof( struct DeviceList), MEMF_PUBLIC|MEMF_CLEAR))
			{
				STRPTR BVolumeName;

				if ( BVolumeName = CStrToBStr( VolumeName))
					{
						struct DosInfo *DosInfo;

						DosInfo = ( struct DosInfo *)BADDR( ( (struct RootNode *)DOSBase->dl_Root)->rn_Info);
						DBG_Info( KPrintF("%s, %s(), %l04d>DosInfo:%l08x\n", __FILE__, __FUNC__, __LINE__, (LONG)DosInfo););

						VolumeNode->dl_Type = DLT_VOLUME;
						VolumeNode->dl_Task = Data->HandlerPort;
						VolumeNode->dl_VolumeDate = *VolumeDate;
						VolumeNode->dl_LockList = NULL;	/* not yet */
						VolumeNode->dl_DiskType = ID_DOS_DISK;
						VolumeNode->dl_Name = (BSTR)MKBADDR( BVolumeName);

						/*	this limitation block me: so I use the hated forbid()/permit
						/*  Because  this  method  is  normaly  invoked by
							another  process than the handler, AddDosEntry() can be
							used.  Otherwise this can deadlock (see autodoc) */
						if ( AddDosEntry( (struct DosList *)VolumeNode))
							{
								Data->VolumeNode = VolumeNode;
								return TRUE;
							}
						FreeVec( VolumeNode);
						*/
						{
							/* In fact, there are no control at all ! */
							VolumeNode->dl_Next = DosInfo->di_DevInfo;
							DosInfo->di_DevInfo = MKBADDR( VolumeNode);
							Permit();
							Data->VolumeNode = VolumeNode;
							return TRUE;
						}
					}
			}
		}
	else
		{
			DBG_Error( KPrintF("%s, %s(), %l04d>E:Forget Volume Name:%l08x\n", __FILE__, __FUNC__, __LINE__, VolumeName););
			return FALSE;
		}

	return FALSE;
}


ULONG DoFHM_Startup( Class *Cl, Object *O, struct TagItem TagP[], struct FileHandlerData * const Data)
{
	DBG_Trace();
	Data->HandlerPort = &( (struct Process *)FindTask( NULL))->pr_MsgPort;
	Data->HandlerPortMask = (1L << Data->HandlerPort->mp_SigBit);
	Data->HostFileSystemReplyPort = CreateMsgPort();
	Data->HostFileSystemReplyPortMask = (1L << Data->HostFileSystemReplyPort->mp_SigBit);
	Data->HostFileSystemPort = ((struct FileLock *)BADDR( Data->RealFileLock))->fl_Task;
	Data->WaitMask = Data->HandlerPortMask | Data->HostFileSystemReplyPortMask;
	DBG_Info( KPrintF("%s, %s(), %l04d>HandlerPort:%l08x\n", __FILE__, __FUNC__, __LINE__, (LONG)Data->HandlerPort););

	return TRUE;
}

ULONG DoFHM_Handler( Class *Cl, Object *O, struct TagItem TagP[], struct FileHandlerData * const Data)
{
	BOOL Quit = FALSE;

	DBG_Trace();

	while( !Quit)
		{
			ULONG Signal;

			Signal = Wait( Data->WaitMask);
			if ( Signal & Data->HandlerPortMask)
				{
					/*	Dos send a Packet to us so handle it ( very simple) */
					struct DosPacket *Packet = NULL;

					{
						struct Message *Msg;
						Msg = GetMsg( Data->HandlerPort);
						Packet = ( struct DosPacket *)Msg->mn_Node.ln_Name;
					}

					/*	Packet are processed asynchronously by method, this mean
						that getting back from method does not imply the packet are done.
						So you can't use 'on stack' parameters for Method.
					*/
					DBG_Info1( KPrintF("%s, %s(), %l04d>Packet Type:%l08d\n", __FILE__, __FUNC__, __LINE__, (LONG)Packet->dp_Type););
					DBG_Info1( KPrintF("%s, %s(), %l04d>Packet Arg1:$%l08x\n", __FILE__, __FUNC__, __LINE__, (LONG)Packet->dp_Arg1););
					DBG_Info1( KPrintF("%s, %s(), %l04d>Packet Arg2:$%l08x\n", __FILE__, __FUNC__, __LINE__, (LONG)Packet->dp_Arg2););
					DBG_Info1( KPrintF("%s, %s(), %l04d>Packet Arg3:$%l08x\n", __FILE__, __FUNC__, __LINE__, (LONG)Packet->dp_Arg3););
					DBG_Info1( KPrintF("%s, %s(), %l04d>Packet Arg4:$%l08x\n", __FILE__, __FUNC__, __LINE__, (LONG)Packet->dp_Arg4););
					DBG_Info1( KPrintF("%s, %s(), %l04d>Packet Arg5:$%l08x\n", __FILE__, __FUNC__, __LINE__, (LONG)Packet->dp_Arg5););
					DBG_Info1( KPrintF("%s, %s(), %l04d>Packet Arg6:$%l08x\n", __FILE__, __FUNC__, __LINE__, (LONG)Packet->dp_Arg6););
					DBG_Info1( KPrintF("%s, %s(), %l04d>Packet Arg7:$%l08x\n", __FILE__, __FUNC__, __LINE__, (LONG)Packet->dp_Arg7););
					{
						BOOL AsyncWorkStarted = FALSE;
						struct fhpDoPacket *fhpDoPacket;

						if ( fhpDoPacket = AllocVec( sizeof( struct fhpDoPacket), MEMF_ANY))
							{
								fhpDoPacket->fhpdp_AsyncWork.MethodID = FHAM_DoPacket;
								fhpDoPacket->fhpdp_AttrList = NULL;
								fhpDoPacket->fhpdp_DosPacket = Packet;

								if ( StartAsyncMethod( O, NULL, fhpDoPacket))
									{
										AsyncWorkStarted = TRUE;
										if ( fhpDoPacket->fhpdp_AsyncWork.fhpaw_State == FHPState_AsyncWorkFinished)
											{
												Quit = ( fhpDoPacket->fhpdp_AsyncWork.fhpaw_Flags & FHPF_QuitHandler);
												#ifdef DEBUG
													if ( fhpDoPacket->fhpdp_AsyncWork.fhpaw_Data)
														{
															/* The data should normaly be free by whose which had use it */
															DBG_Error( KPrintF("%s, %s(), %l04d>E:fhpDoPacket->fhpdp_Data:%l08x\n", __FILE__, __FUNC__, __LINE__, fhpDoPacket->fhpdp_AsyncWork.fhpaw_Data););
														}
												#endif
												DoMethod( O, FHM_ReplyPacket, ITEMIGNORED, FHP_Packet, fhpDoPacket->fhpdp_DosPacket);
												FreeVec( fhpDoPacket);
											}
											//{
											//	/* an error occur, just reply the packet. The FHM_DoPacket
											//	 should have filled the Res1/Res2 */
											//	DoMethod( O, FHM_ReplyPacket, ITEMIGNORED, FHP_Packet, Packet);
											//}
									}
							}
						if ( !AsyncWorkStarted )
							{
								FreeVec( fhpDoPacket);	/* NULL is valid */
								Packet->dp_Res1 = DOSFALSE;
								Packet->dp_Res2 = ERROR_NO_FREE_STORE;
								DoMethod( O, FHM_ReplyPacket, ITEMIGNORED, FHP_Packet, Packet);
							}
					}
				}
			else if	( Signal & Data->HostFileSystemReplyPortMask)
				{
					/* The Host FileSystem just reply a packet we previously send to it so process it ( continue to process the initial Paquet recieved from DOS) */
					struct DosPacket *Packet = NULL;
					{
						struct Message *Msg;
						Msg = GetMsg( Data->HostFileSystemReplyPort);
						Packet = ( struct DosPacket *)Msg->mn_Node.ln_Name;
						#ifdef DEBUG
							if ( !Msg || !Packet)
								DBG_Error( KPrintF("%s, %s(), %l04d>E:Msg:%l08x Packet:%l08x\n", __FILE__, __FUNC__, __LINE__, Msg, Packet););
						#endif
					}
					if ( Packet == Data->ServicePacketInProgress) /* Verify that the Packet replyed are those we send */
						{
							struct DosPacket *ServicePacket;
							struct fhpAsyncWork *fhpAsyncWork;

							ServicePacket = Data->ServicePacketInProgress;
							fhpAsyncWork = Data->fhpAsyncWorkInProgress;
							/* don't forget this or FHM_SendAWaitingServicePacket() don't do anything */
							Data->ServicePacketInProgress = NULL;
							Data->fhpAsyncWorkInProgress = NULL;

							/* Send an other ServicePacket which wait to be send. Send it the most early possible so we can take advantage of asynchronous processing */
							DoMethod( O,
								FHM_SendAWaitingServicePacket, ITEMIGNORED,
								TAG_END);
				
							/* now continue to process the Packet which cause this ServicePacket to HostFS */
							{
								struct fhpDoPacket *fhpDoPacket;

								if ( fhpDoPacket = (struct fhpDoPacket *)ContinueAsyncMethod( O, fhpAsyncWork))
									{
										/* ContinueAsyncMethod() garentued that ALL stacked work are finished */
										if ( fhpDoPacket->fhpdp_AsyncWork.fhpaw_State == FHPState_AsyncWorkFinished)
											{
												Quit = ( fhpDoPacket->fhpdp_AsyncWork.fhpaw_Flags & FHPF_QuitHandler);
												#ifdef DEBUG
													if ( fhpDoPacket->fhpdp_AsyncWork.fhpaw_Data)
														{
															/* The data should normaly be free by whose which had use it */
															DBG_Error( KPrintF("%s, %s(), %l04d>E:fhpDoPacket->fhpdp_Data:%l08x\n", __FILE__, __FUNC__, __LINE__, fhpDoPacket->fhpdp_AsyncWork.fhpaw_Data););
														}
												#endif
												DoMethod( O, FHM_ReplyPacket, ITEMIGNORED, FHP_Packet, fhpDoPacket->fhpdp_DosPacket);
												FreeVec( fhpDoPacket);
											}
										else
											{
												DBG_Error( KPrintF("%s, %s(), %l04d>E:fhpDoPacket not realy finished\n", __FILE__, __FUNC__, __LINE__););
											}
									}
							}
						}
					else
						{
							/* Panic ! */
							Quit = TRUE;
							DBG_Error( KPrintF("%s, %s(), %l04d>E:Packet replyed by HFS are not those sended. Sended:%l08x, Replyed:%l08x\n", __FILE__, __FUNC__, __LINE__, Packet, Data->ServicePacketInProgress););
						}
				}
			else 
				{
					/* subclass signal, so call warn it...*/
					/* not yet tmp*/
				}
		}

	return TRUE;
}


ULONG DoFHM_HandleSignal( Class *Cl, Object *O, struct TagItem TagP[], struct FileHandlerData * const Data)
{
	ULONG RC = FALSE;	/* return say if we continue/stop */
	ULONG Signal;

	DBG_Trace();

	Signal = GetTagData( FHP_Signal, 0, TagP);
	DBG_Info( KPrintF("%s, %s(), %l04d>Signal:%l08x\n", __FILE__, __FUNC__, __LINE__, Signal););

	return RC;
}


ULONG DoFHM_NewFileLock( Class *Cl, Object *O, struct TagItem TagP[], struct FileHandlerData * const Data)
{
	LONG Access;
	void *flData;

	struct FileLock *FileLock;

	Access = (LONG)GetTagData( FHP_Access, SHARED_LOCK, TagP);
	flData = (void *)GetTagData( FHP_Data, NULL, TagP);
	if ( FileLock = AllocVec( sizeof( struct FileLock), MEMF_PUBLIC))
		{
			FileLock->fl_Link = NULL;
			FileLock->fl_Key = (LONG)flData;
			FileLock->fl_Access = Access;
			FileLock->fl_Task = Data->HandlerPort;
			FileLock->fl_Volume = MKBADDR( Data->VolumeNode);

			return (ULONG)FileLock;
		}

	return NULL;
}


#include "proto/Handler.h"

ULONG DoFHM_MountDOSDev( Class *Cl, Object *O, struct TagItem TagP[], struct FileHandlerData * const Data)
{
	ULONG RC = FALSE;
	STRPTR DOSDevName;
	struct Process *HandlerProcess;

	DBG_Trace();
	DOSDevName = (STRPTR)GetTagData( FHP_StrPtr, NULL, TagP);

	HandlerProcess = CreateNewProcTags(
		NP_Entry, HandlerEntry,
		NP_StackSize, 14096,
		NP_Name, DOSDevName,
		NP_Priority, 4,	/* dbg */
		TAG_END);
	if ( HandlerProcess)
		{
			struct DeviceNode *DeviceNode;

			if ( DeviceNode = AllocVec( sizeof( struct DeviceNode), MEMF_PUBLIC))
				{
					STRPTR BDOSDevName;

					if ( BDOSDevName = CStrToBStr( DOSDevName))
						{
		
							DeviceNode->dn_Next = NULL;
							DeviceNode->dn_Type = DLT_DEVICE;
							DeviceNode->dn_Task = &HandlerProcess->pr_MsgPort;
							DeviceNode->dn_Lock = NULL;
							DeviceNode->dn_Handler = MKBADDR( "No file !");
							DeviceNode->dn_StackSize = 0;	/* alredy provided when created the handlerProcess */
							DeviceNode->dn_Priority = 0;	/* alredy provided when created the handlerProcess */
							DeviceNode->dn_Startup = MKBADDR( O);	/* somewhat redondency, but seems a good place for it */
							DeviceNode->dn_SegList = NULL;
							DeviceNode->dn_GlobalVec = -1;	/* normaly not used: HandlerProcess alredy created */
							DeviceNode->dn_Name = MKBADDR( BDOSDevName);
							{
								struct Library *ExpansionBase = OpenLibrary( "expansion.library", 37L);

								if ( AddDosNode( -128, ADNF_STARTPROC, DeviceNode))
									{
										/* Device are now mounted */
										DBG_Info1( KPrintF("%s, %s(), %l04d>Mounted\n", __FILE__, __FUNC__, __LINE__););
										Data->DevNode = DeviceNode;
										HandlerProcess->pr_Task.tc_UserData = O;	/* put the Object where the Handler Process ca retive it */

										RC = TRUE;
									}
								CloseLibrary( ExpansionBase);
						}
					}
				}
			Signal( &HandlerProcess->pr_Task, SIGBREAKF_CTRL_D);	/* and signal the HandlerProcess it can retrive the Object */
		}

	return FALSE;
}

