/*
 *    Copyright  1994 by S.R.
 *
 *      Created: 17 May 1994
 *      Modified:	08 Jun 1994	19:14:15
 *
 * Make>> smake
 */

//#undef DEBUG
#include "/Debug.h"

#include "/File-HandlerClass.h"
#include "/File-Handler_pragmas.h"
#include "/File-Handler_protos.h"

#include "/Tar-FileHandlerClass.h"
#include "Tar-FileHandler.h"

#include "TarObject.h"

#include "proto/TarObject.h"
#include "proto/Misc.h"

#define TFHM_GO_Private1 TFHM_GO_LoadDir

extern struct Library *FileHandlerBase;

#define NULLNAME( n) ( n[ 0] == '\0')
#define ISSAMESIGN( a, b) ( ( a * b) >= 0 )

#define ANYTYPE 0		/* any ST_ */


#define TARHEADERSIZE 512	/* not the size of the struct, but this still the size of the header ! */
#define NAMESIZE 100
#define TUNAMELEN 32
#define TGNAMELEN 32

struct TarHeader
{
	char FileName[NAMESIZE];
	char Mode[8];
	char UID[8];
	char GID[8];
	char Size[12];		/* in octal  */
	char mtime[12];
	char CheckSum[8];
	char LinkFlag;
	char Archive_LinkName[NAMESIZE];
	char Magic[8];
	char UserName[TUNAMELEN];
	char GroupName[TGNAMELEN];
	char DevMajor[8];
	char DevMinor[8];
};


static struct TarObject *NewTarObject(struct TarObject *ParentTarObject, ULONG Type, LONG Access, char Name[], ULONG Size, ULONG Offset)
{
	struct TarObject *TarObject;

	if (TarObject = AllocVec(sizeof (struct TarObject), MEMF_ANY))
	{
		TarObject->Type = Type;
		NewMinList(&TarObject->TarObjectList);
		TarObject->ParentTarObject = ParentTarObject;
		TarObject->Access = Access;
		TarObject->AccessCnt = 0;
		if (Name)
		{
			UBYTE l;

			strcpy(TarObject->Name, Name);
			l = strlen(TarObject->Name) - 1;
			if (TarObject->Name[l] == '/')
				TarObject->Name[l] = '\0';
		}
		else
			strcpy(TarObject->Name, "Unnamed");
		TarObject->Size = Size;
		TarObject->Offset = Offset;

		return TarObject;
	}

	return NULL;
}


static struct TarObject *GetLocalTarObject(struct TarObject *ParentTarObject, char Name[], LONG Type)
{
	struct TarObject *i;

	for (i = (struct TarObject *)GetHead(&ParentTarObject->TarObjectList); i; i = (struct TarObject *)GetSucc(i))
	{
		if (Stricmp(Name, i->Name) == 0)
		{
			if (!Type || (i->Type == Type))
				return i;
		}
	}

	return NULL;
}


static struct TarObject *GetTarObject(struct TarObject *ParentTarObject, STRPTR FullPathName, LONG Type)
{
	char *p;
	char Part[108];
	UBYTE i;

	for (i = 0, p = PartName(FullPathName, Part, 0, FALSE); p && !NULLNAME(p); p = PartName(FullPathName, Part, ++i, FALSE))
	{
		if (!(ParentTarObject = GetLocalTarObject(ParentTarObject, p, Type)))
		{
			/* the dir/file does not exist in this dir (may be an error of the usr */
			return NULL;
		}
	}

	return ParentTarObject;	/* not the parent but the object itself */
}


ULONG DoTFHAM_ReadAllDir(Class * Cl, Object * O, struct fhpAsyncWork * fhpAsyncWork, struct TarFileHandlerData * const Data)
{

	switch (fhpAsyncWork->fhpaw_State)
	{
		case FHPState_Initial:
		{
			if (Data->Root = NewTarObject(NULL, ST_ROOT, 0, "TarFileVolume", 0, 0))
			{
				struct fhpAsyncRW *fhpAsyncRW;
				/* tmp: 2* */
				if (fhpAsyncRW = AllocVec(sizeof (struct fhpAsyncRW) + 2 * TARHEADERSIZE, MEMF_PUBLIC))		/* MEMF_PUBLIC because it is a buffer memory wich can be shared between multiple process */
				{
					fhpAsyncWork->fhpaw_Data = fhpAsyncRW;
					fhpAsyncRW->fhparw_AsyncWork.MethodID = FHAM_AsyncRead;
					fhpAsyncRW->fhparw_Pos = 0;
					fhpAsyncRW->fhparw_Len = TARHEADERSIZE;
					fhpAsyncRW->fhparw_Mem = (UBYTE *) (fhpAsyncRW + sizeof (fhpAsyncRW));
					if (StartAsyncMethod(O, fhpAsyncWork, fhpAsyncRW))
					{
						return TRUE;
					}
					FreeVec(fhpAsyncRW);
				}
			}
			AMReturn(fhpAsyncWork, FALSE, 0, FALSE, 1);
		}
		break;
		case FHPState_ReturnFromHostFS:
		{
			struct fhpAsyncRW *fhpAsyncRW;

			fhpAsyncRW = fhpAsyncWork->fhpaw_Data;
			fhpAsyncWork->fhpaw_Data = NULL;

			if (fhpAsyncRW->fhparw_AsyncWork.fhpaw_RC != -1)
			{
				struct TarHeader *TarHeader;

				TarHeader = (struct TarHeader *)fhpAsyncRW->fhparw_Mem;
				if (TarHeader->FileName[0])
				{
					ULONG Size;
					ULONG NextHeaderOffset;
					struct TarObject *TarObject;
					struct TarObject *ParentTarObject;
					char BaseName[108];
					char PathName[108];
					ULONG Type;
					ULONG Offset;

					Type = IsDir(TarHeader->FileName) ? ST_USERDIR : ST_FILE;
					Size = Oct(TarHeader->Size);
					DivideName(TarHeader->FileName, PathName, BaseName);
					if (!(ParentTarObject = GetTarObject(Data->Root, PathName, ST_USERDIR)))
						ParentTarObject = Data->Root;	/* the tar format garented that parent dir had been reach before is contents, so the parent dir must be the root  */

					Offset = fhpAsyncRW->fhparw_Pos - fhpAsyncRW->fhparw_Len + TARHEADERSIZE;
					if (ParentTarObject && (TarObject = NewTarObject(ParentTarObject, Type, 0, BaseName, Size, Offset)))
					{
						AddTail((struct List *)&ParentTarObject->TarObjectList, (struct Node *)TarObject);
						NextHeaderOffset = NumberOfBlock(TarObject->Size, 512) * 512;	/* round up to higher block */

						fhpAsyncWork->fhpaw_Data = fhpAsyncRW;
						fhpAsyncRW->fhparw_AsyncWork.MethodID = FHAM_AsyncRead;
						fhpAsyncRW->fhparw_Pos += NextHeaderOffset;	/* this make OFFSET_CURRENT's like position */
						fhpAsyncRW->fhparw_Len = TARHEADERSIZE;
						fhpAsyncRW->fhparw_Mem = (UBYTE *) (fhpAsyncRW + sizeof (fhpAsyncRW));
						if (StartAsyncMethod(O, fhpAsyncWork, fhpAsyncRW))
						{
							return TRUE;
						}
						FreeVec(fhpAsyncRW);
						AMReturn(fhpAsyncWork, FALSE, 0, FALSE, 1);
					}
				}
				else
				{
					/* scan of the tar file are finished */
					FreeVec(fhpAsyncRW);
					AMReturn(fhpAsyncWork, TRUE, 0, TRUE, 0);
				}
			}
		}
		break;
	default:
		{
			DBG_Error(KPrintF("%s, %s(), %l04d>E:\0x07Unknown fhpAsyncWork state:%ld\n", __FILE__, __FUNC__, __LINE__, (LONG) fhpAsyncWork->fhpaw_State));
			fhpAsyncWork->fhpaw_State = FHPState_AsyncWorkFinished;
			return FALSE;
		}
		break;
	}

}



struct TarObject *GetTarObjectFromParentLock(struct FileLock *ParentFileLock, char BName[], struct TarFileHandlerData *const Data)
{
	struct TarObject *ParentTarObject = NULL;
	char Name[108];

	if (ParentFileLock)
		ParentTarObject = (struct TarObject *)ParentFileLock->fl_Key;
	else
		ParentTarObject = Data->Root;

	if (BName[0] == 0)
		return ParentTarObject;

	strncpy(Name, &BName[1], BName[0]);
	Name[BName[0]] = '\0';
	return GetTarObject(ParentTarObject, Name, ANYTYPE);
}
