/*
#define DEBUG
*/

#include "LHAutils.h"
LHAEntry	*Alloc_LHA_Entry(LzHeader *header,char *name,LONG type)
{
LHAEntry	*entry;

	entry = (LHAEntry *)AllocMem(sizeof(LHAEntry),MEMF_PUBLIC|MEMF_CLEAR);
	if (entry==NULL) return NULL;
	entry->DirEntryType = type;
	entry->brothers.ln_Name = (char *)AllocMem(strlen(name)+1,
		MEMF_PUBLIC|MEMF_CLEAR);
	if (entry->brothers.ln_Name==NULL) {
		FreeMem(entry,sizeof(LHAEntry));
		return NULL;
	}
	strcpy(entry->brothers.ln_Name,name);
#ifdef DEBUG
	printf("DEBUG Create entry [%s]\n",entry->brothers.ln_Name);
#endif /* DEBUG */
	entry->Protection = 0L;
	entry->Size = header->original_size;
	entry->DateStamp = header->last_modified_stamp;
	switch(type) {
	case ST_PIPEFILE :
	case ST_LINKFILE :
	case ST_FILE :
		entry->sons = NULL;
		break;
	case ST_ROOT :
	case ST_USERDIR :
	case ST_SOFTLINK :
	case ST_LINKDIR :
		entry->sons = (struct List *)AllocMem(sizeof(struct List),
			MEMF_PUBLIC|MEMF_CLEAR);
		if (entry->sons==NULL) {
			FreeMem(entry->brothers.ln_Name,
				strlen(entry->brothers.ln_Name)+1);
			FreeMem(entry,sizeof(LHAEntry));
			return NULL;
		}
		NewList(entry->sons);
		break;
	}
	return entry;
}

ERROR	Add_Header_2_LHA_Tree(LzHeader *header,struct List *tree)
{
int	indice;
char	*name, ex_part[MAX_EXTRACTION];
LHAEntry	*entry;
struct List	*current;
struct Node	*found;

#ifdef DEBUG
	printf("Add_Header_2_LHA_Tree [%s]\n",header->name);
#endif
	current = tree;
	name = header->name;
	indice = 0;
	while (name[indice]!='\0') {
		indice = SplitName(name,DELIM,ex_part,indice,MAX_EXTRACTION);
		if ((found=FindName(current,ex_part))==NULL) {
			entry = Alloc_LHA_Entry(header,ex_part,
					indice==-1?ST_FILE:ST_USERDIR);
			if (entry==NULL) return BAD_ALLOC;
			found = (struct Node *)entry;
			AddTail(current,found);
		}
		if (indice==-1) break;
		current = ((LHAEntry *)found)->sons;
	}
	return NO_ERROR;
}

void	Print_LHAEntry(LHAEntry *entry,char *dir_name)
{
	switch(entry->DirEntryType) {
		case ST_PIPEFILE: printf("pipe file "); break;
		case ST_LINKFILE: printf("link file "); break;
		case ST_FILE: printf("%9d ",entry->Size); break;
		case ST_ROOT: printf("root dir  "); break;
		case ST_USERDIR: printf("directory "); break;
		case ST_SOFTLINK: printf("soft link "); break;
		case ST_LINKDIR: printf("link dir  "); break;
		default: printf("unk type  \n"); break;
	}
	printf("%s%s\n",dir_name,entry->brothers.ln_Name);
}

void	Print_LHA_Tree(struct List *root,char *path)
{
char	buffer[100];
struct Node	*node;
LHAEntry	*entry;

	if (IsListEmpty(root)) return;
	strcpy(buffer,path);
	/* On parcourt d'abord la liste des fils */
	for (node = root->lh_Head; node->ln_Succ; node = node->ln_Succ) {
		entry = (LHAEntry *)node;
		if (entry->DirEntryType==ST_PIPEFILE ||
		    entry->DirEntryType==ST_LINKFILE ||
		    entry->DirEntryType==ST_FILE)
			Print_LHAEntry(entry,buffer);
	}
	/* On descends dans la hierarchie */
	for (node = root->lh_Head; node->ln_Succ; node = node->ln_Succ) {
		entry = (LHAEntry *)node;
		if (entry->DirEntryType==ST_ROOT ||
		    entry->DirEntryType==ST_USERDIR ||
		    entry->DirEntryType==ST_SOFTLINK ||
		    entry->DirEntryType==ST_LINKDIR) {
			Print_LHAEntry(entry,buffer);
			strcat(buffer,entry->brothers.ln_Name);
			strcat(buffer,"/");
			Print_LHA_Tree(entry->sons,buffer);
		}
	}
}

void	Free_LHA_Tree(struct List *root)
{
LHAEntry	*entry;

	if (IsListEmpty(root)) return;
	while ((entry=(LHAEntry *)RemHead(root))!=NULL) {
		switch (entry->DirEntryType) {
		case ST_ROOT :
		case ST_USERDIR :
		case ST_SOFTLINK :
		case ST_LINKDIR :
			Free_LHA_Tree(entry->sons);
			FreeMem(entry->sons,sizeof(struct List));
		case ST_PIPEFILE :
		case ST_LINKFILE :
		case ST_FILE :
			FreeMem(entry->brothers.ln_Name,
				strlen(entry->brothers.ln_Name)+1);
			FreeMem(entry,sizeof(LHAEntry));
			break;
		}
	}
}
