/*
 *  Alloc.c - Copyright  1994 by S.R. & P.C.
 *
 *  Created:    21 Nov 1994  23:47:38
 *  Modified:   03 Dec 1994  23:11:49
 *
 *  Make>> smake
 */


#include <clib.h>
#include <archives/ArchiveClass.h>

#include "Archive.h"
#include "Debug.h"

#include "proto/Alloc.h"
#include "proto/Misc.h"

extern struct ArchiveBase *ArchiveBase;


static APTR AllocVecPool(struct ArchiveData *Data, ULONG Size)
{
	ULONG *Chunk;

	// Protect pool against concurrent access (handler / slave process)
	ObtainSemaphore(&Data->MemPoolSemaphore);
	Size += 4;
	if (Chunk = LibAllocPooled(Data->MemoryPool, Size)) {
		*Chunk = Size;
		Chunk++;
	}
	ReleaseSemaphore(&Data->MemPoolSemaphore);
	return Chunk;
}


static void FreeVecPool(struct ArchiveData *Data, APTR Chunk)
{
	ULONG *Mem = Chunk;

	if (Mem) {
		Mem--;
		ObtainSemaphore(&Data->MemPoolSemaphore);
		LibFreePooled(Data->MemoryPool, Mem, *Mem);
		ReleaseSemaphore(&Data->MemPoolSemaphore);
	}
}


APTR __saveds __asm AllocArchiveObjectA(register __a0 Object *O, register __d0 ULONG Type, register __a1 struct TagItem *TagList)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	struct ArchiveClassData *acd;
	ULONG *Chunk = NULL, Size, AllocSize;

	switch (Type) {
	case AO_Custom:
		if (Size = GetTagData(AAO_ObjectSize, 0, TagList))
			Chunk = AllocVecPool(Data, Size);
		break;
	case AO_ASyncRequest:
		// ASyncRequests are standard messages with hidden StackSwapStruct BEFORE message.
		// This way, it may work with any OS exec messages extension such IORequests...
		Size = GetTagData(AAO_ObjectSize, sizeof(struct ASyncRequest), TagList);
		AllocSize = Size + sizeof(struct StackSwapStruct);			// + 12
		if (Chunk = AllocVecPool(Data, AllocSize)) {
			struct StackSwapStruct *ss;
			struct ASyncRequest *M;

			ss = (struct StackSwapStruct *)Chunk;
			if (ss->stk_Lower = AllocVec(STACKSIZE, MEMF_PUBLIC)) {
				ss->stk_Upper = (ULONG)ss->stk_Lower + STACKSIZE;
				ss->stk_Pointer = (APTR)(ss->stk_Upper - 4);	// - 4 for safety
				Chunk += 3;
				M = (struct ASyncRequest *)Chunk;
				M->mn_ReplyPort = Data->ASyncReplyPort;
				M->mn_Length = Size;
				M->mn_Node.ln_Type = NT_MESSAGE;
			}
			else {
				FreeVecPool(Data, Chunk);
				Chunk = NULL;
			}
		}
		break;
	case AO_Lock:
		Chunk = AllocVec(sizeof(struct FileLock), MEMF_PUBLIC|MEMF_CLEAR);	// Std mem alloc for locks
		break;
	case AO_Entry:
		acd = (struct ArchiveClassData *)OCLASS(O)->cl_UserData;
		Chunk = AllocVecPool(Data, acd->EntryOffset + acd->EntrySize);
		NewMinList(&((struct PrivateArchiveEntry *)Chunk)->ae_SubEntries);
		break;
	case AO_Handle:
		acd = (struct ArchiveClassData *)OCLASS(O)->cl_UserData;
		Chunk = AllocVecPool(Data, acd->HandleOffset + acd->HandleSize);
		break;
	}
	return Chunk;
}


void __saveds __asm FreeArchiveObject(register __a0 Object *O, register __d0 ULONG Type, register __a1 APTR Chunk)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	long *Mem = Chunk;

	switch (Type) {
	case AO_Lock:
		FreeVec(Chunk);
		break;
	case AO_ASyncRequest:
		{
			struct StackSwapStruct *ss;

			Mem -= 3;
			ss = (struct StackSwapStruct *)Mem;
			FreeVec(ss->stk_Lower);
		}
	default:
		FreeVecPool(Data, Mem);
	}
}

