/*
 *  Archive.c - Copyright  1994 by S.R. & P.C.
 *
 *  Created:    25 Jul 1994
 *  Modified:   07 Dec 1994  12:53:44
 *
 *  Make>> smake
 */

#include <clib.h>
#include <archives/ArchiveClass.h>
#include <libraries/WhatIsBase.h>
#include <proto/WhatIs.h>

#include "Archive.h"
#include "Debug.h"

#include "proto/Archive.h"
#include "proto/Handler.h"
#include "proto/ASyncStuff.h"
#include "proto/Misc.h"

extern struct ArchiveBase *ArchiveBase;

#include <archives/ArchiveMethods.h>

#define ARCHIVE_VERSION		1		// Version of subclass libraries required


struct ArchiveNode {
	struct MinNode Node;
	struct ArchiveType ArchiveType;
	ULONG FileType;
};


void FreeArchiveTypes(void)
{
	APTR ArchiveNode;

	while (ArchiveNode = RemTail((struct List *)&ArchiveBase->ArchiveTypes))
		FreeVec(ArchiveNode);
}


LONG __saveds __asm AddArchiveTypeA(register __a0 UBYTE *FileName, register __a1 struct TagItem *TagList)
{
	struct ArchiveNode *ArchiveNode;
	struct Library *IFFParseBase;
	struct ContextNode *top;
	struct IFFHandle *iff;
	LONG Ok = FALSE;

	if (!(IFFParseBase = OpenLibrary("iffparse.library", 37)))
		return FALSE;
	if (ArchiveNode = AllocVec(sizeof(struct ArchiveNode), MEMF_PUBLIC|MEMF_CLEAR)) {
		if (iff = AllocIFF()) {
			if (iff->iff_Stream = Open(FileName, MODE_OLDFILE)) {
				InitIFFasDOS(iff);
				if (OpenIFF(iff, IFFF_READ) == 0) {
					StopChunk(iff, ID_ARCV, ID_ARCT);
					while (!ParseIFF(iff, IFFPARSE_SCAN)) {
						if (!(top = CurrentChunk(iff)))
							break;
						if (top->cn_ID == ID_ARCT) {
							if (top->cn_Size == sizeof(struct ArchiveType)) {
								if (ReadChunkBytes(iff, &ArchiveNode->ArchiveType, top->cn_Size) == top->cn_Size) {
									ArchiveNode->FileType = GetIDType(ArchiveNode->FileTypeName);
									if (ArchiveNode->FileType != TYPE_UNKNOWNIDSTRING) {
										AddTail((struct List *)&ArchiveBase->ArchiveTypes, (struct Node *)ArchiveNode);
										Ok = TRUE;
									}
								}
							}
							break;	// Only one chunk for now.
						}
					}
					CloseIFF(iff);
				}
				Close(iff->iff_Stream);
			}
			FreeIFF(iff);
		}
		CloseLibrary(IFFParseBase);
		if (!Ok)
			FreeVec(ArchiveNode);
	}
	return Ok;
}


void LoadArchiveTypes(void)
{
	#define MAX_PATHLEN 256
	struct AnchorPath *AP;
	LONG MatchErr;
	BPTR CD, ArcDir;
	
	if (ArcDir = Lock("DEVS:Archives", ACCESS_READ)) {
		CD = CurrentDir(ArcDir);
		if (AP = AllocVec(sizeof(struct AnchorPath) + MAX_PATHLEN, MEMF_ANY|MEMF_CLEAR)) {
			AP->ap_Strlen = MAX_PATHLEN;
			for (MatchErr = MatchFirst("~(#?.info)", AP); MatchErr == 0; MatchErr = MatchNext(AP)) {
				if (AP->ap_Info.fib_DirEntryType < 0)
					AddArchiveTypeA(AP->ap_Info.fib_FileName, NULL);
			}
			FreeVec(AP);
		}
		CurrentDir(CD);
		UnLock(ArcDir);
	}
}


static char *GetArchiveBaseName(UBYTE *FileName, struct TagItem *TagList)
{
	struct ArchiveNode *ArchiveNode;
	ULONG Type;
	
	if (!(Type = GetTagData(NA_FileType, 0, TagList))) {
		__aligned struct FileInfoBlock fib;
		BPTR FL, Dir, CD;
	
		if (FL = Lock(FileName, ACCESS_READ)) {
			if (Dir = ParentDir(FL)) {
				CD = CurrentDir(Dir);
				Examine(FL, &fib);
				Type = WhatIsTags(fib.fib_FileName, WI_FIB, &fib, WI_Deep, DEEPTYPE, TAG_END);
				CurrentDir(CD);
				UnLock(Dir);
			}
			UnLock(FL);
		}
		else
			SetIoErr(ERROR_OBJECT_NOT_FOUND);
	}
	for (ArchiveNode = GetHead(&ArchiveBase->ArchiveTypes); ArchiveNode; ArchiveNode = GetSucc(ArchiveNode)) {
		if (!CmpFileType(Type, ArchiveNode->FileType))
			return ArchiveNode->ObjectBaseName;
	}
	SetIoErr(ERROR_OBJECT_WRONG_TYPE);
	return NULL;
}


Object * __saveds __asm NewArchiveA(register __a0 UBYTE *FileName, register __a1 struct TagItem *TagList)
{
	char *ArchiveBaseName, ObjectLibName[40], ObjectClassName[40];
	struct Library *ObjLibBase;
	Object *ArchiveObject;

	DBG_Trace();
	if (!(ArchiveBaseName = GetArchiveBaseName(FileName, TagList)))
		return NULL;
	SPrintf(ObjectLibName, "archives/%s.archive", ArchiveBaseName);
	SPrintf(ObjectClassName, "%s.archive", ArchiveBaseName);
	if (!(ObjLibBase = OpenLibrary(ObjectLibName, ARCHIVE_VERSION)))
		return NULL;
	if (ArchiveObject = NewObject(NULL, ObjectClassName, AA_ArchiveFileName, FileName, TAG_MORE, TagList)) {
		struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, ArchiveObject);

		Data->ObjLibBase = ObjLibBase;
	}
	else {
		CloseLibrary(ObjLibBase);
		SetIoErr(ERROR_NO_FREE_STORE);
	}
	return ArchiveObject;
}


void __saveds __asm DisposeArchive(register __a0 Object *O)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	struct Library *Lib;

	Lib = Data->ObjLibBase;		// Keep a copy before dispose!!
	DisposeObject(O);
	CloseLibrary(Lib);
}


#define MEMPOOL_SIZE	(8*1024)
#define MEMPOOL_THRES	(4*1024)

static ULONG OM_New(Class *Cl, Object *O, struct opSet const *oS)
{
	ULONG RC = NULL;			/* generic return value, meaning depend on the method */
	struct opSet opSet = *oS;
	APTR Pool;

	DBG_Trace();
	if (!(Pool = LibCreatePool(MEMF_PUBLIC|MEMF_CLEAR, MEMPOOL_SIZE, MEMPOOL_THRES)))
		return NULL;
	if (RC = DoSuperMethodA(Cl, O, (Msg) &opSet)) {
		const Object *ThisObject = (Object *)RC;
		struct ArchiveData *const Data = INST_DATA(Cl, ThisObject);

		memset(Data, 0, sizeof(struct ArchiveData));
		Data->MemoryPool = Pool;
		InitSemaphore(&Data->MemPoolSemaphore);
	}
	return RC;
}


static ULONG OM_Dispose(Class *Cl, Object *O, Msg M)
{
	struct ArchiveData *const Data = INST_DATA(Cl, O);

	DBG_Trace();
	if (Data->MemoryPool)
		LibDeletePool(Data->MemoryPool);
	return DoSuperMethodA(Cl, O, M);
}


static ULONG OM_Set(Class *Cl, Object *O, struct opSet * const oS)
{
	return DoSuperMethodA(Cl, O, (Msg) oS);
}


static ULONG OM_Get(Class *Cl, Object *O, struct opGet *const oG)
{
	struct ArchiveData *const Data = INST_DATA(Cl, O);
	ULONG RC = 0;

	switch (oG->opg_AttrID) {
	case AA_DeviceBaseName:
		*oG->opg_Storage = (ULONG)"DDV";	// Dummy value. Subclasses may intercept this and set its own
		break;
	case AA_DeviceName:
		*oG->opg_Storage = (ULONG)Data->DeviceName;
		break;
	case AA_VolumeName:
		*oG->opg_Storage = (ULONG)"VOL";	// Dummy value. Subclasses may intercept this and set its own
		break;
	default:
		RC = DoSuperMethodA(Cl, O, (Msg) oG);
	}
	return RC;
}


static ULONG __saveds __asm far Dispatch(register __a0 Class *Cl, register __a2 Object *O, register __a1 Msg M)
{
	ULONG RC = NULL;	/* generic return value, meaning depend on the method */

	if ((M->MethodID & AAHM_Dummy) == AAHM_Dummy) {
		#ifdef DEBUG
		if ((struct Process *)SysBase->ThisTask == Data->HandlerProcess) {
			SimpleReq("ERROR! ERROR! ERROR! ERROR! ERROR!\n!!!! ASynchronous method called in handler context !!!!");
			return 0;
		}
		#endif	
		switch (M->MethodID) {
		case AAHM_ScanArchive:
			aahm_ScanArchive(Cl, O, (struct opTagList *)M);
			break;
		case AAHM_WaitEndScan:	// Dummy method. Just use async mechanism to wait for end of scan
			break;
		}
	}
	else {
		switch (M->MethodID) {
		case AHM_DoPacket:
			DBG_Info1(KPrintF("%s, %s(), %04ld> AHM_DoPacket\n", __FILE__, __FUNC__, __LINE__));
			RC = ahm_DoPacket(Cl, O, (struct ahpDoPacket *)M);
			break;
		case OM_GET:
			DBG_Info1(KPrintF("%s, %s(), %04ld> OM_GET\n", __FILE__, __FUNC__, __LINE__));
			RC = OM_Get(Cl, O, (struct opGet * const)M);
			break;
		case OM_SET:
			DBG_Info1(KPrintF("%s, %s(), %04ld> OM_SET\n", __FILE__, __FUNC__, __LINE__));
			RC = OM_Set(Cl, O, (struct opSet * const)M);
			break;
		case OM_NEW:
			DBG_Info1(KPrintF("%s, %s(), %04ld> OM_NEW\n", __FILE__, __FUNC__, __LINE__));
			RC = OM_New(Cl, O, (struct opSet * const)M);
			break;
		case OM_DISPOSE:
			DBG_Info1(KPrintF("%s, %s(), %04ld> OM_DISPOSE\n", __FILE__, __FUNC__, __LINE__));
			RC = OM_Dispose(Cl, O, M);
			break;
		case AM_Mount:
			DBG_Info1(KPrintF("%s, %s(), %04ld> AM_Mount\n", __FILE__, __FUNC__, __LINE__));
			RC = am_Mount(Cl, O, (struct opTagList *)M);
			break;
		case AHM_Startup:
			DBG_Info1(KPrintF("%s, %s(), %04ld> AHM_Startup\n", __FILE__, __FUNC__, __LINE__));
			RC = ahm_Startup(Cl, O, (struct opTagList *)M);
			break;
		case AHM_Finish:
			DBG_Info1(KPrintF("%s, %s(), %04ld> AHM_Finish\n", __FILE__, __FUNC__, __LINE__));
			ahm_Finish(Cl, O, (struct opTagList *)M);
			break;
		default:
			DBG_Info1(KPrintF("%s, %s(), %04ld> MethodID:$%08lx\n", __FILE__, __FUNC__, __LINE__, M->MethodID));
			RC = DoSuperMethodA(Cl, O, M);
			break;
		}
	}
	DBG_Info1(KPrintF("%s, %s(), %l04d> Return:$%l04x\n", __FILE__, __FUNC__, __LINE__, RC));
	return RC;
}


Class *InitCl(void)
{
	Class *Cl;

	DBG_Trace();
	if (Cl = MakeClass(ARCHIVECLASS, "rootclass", NULL, sizeof(struct ArchiveData), 0)) {
		static struct ArchiveClassData acd;

		acd.EntrySize = sizeof(struct PrivateArchiveEntry);
		acd.HandleSize = sizeof(struct PrivateArchiveHandle);
		Cl->cl_UserData = (ULONG)&acd;
		Cl->cl_Dispatcher.h_Entry = (ULONG(*)())Dispatch;
		Cl->cl_Dispatcher.h_SubEntry = NULL;
		//Cl->cl_Dispatcher.h_Data = &ArchiveData;  /* not neccsary in a Class */
		AddClass(Cl);
	}
	return Cl;
}


BOOL FreeCl(Class *Cl)
{
	DBG_Trace();
	if (Cl) {
		RemoveClass(Cl);
		return FreeClass(Cl);
	}
	return TRUE;
}


LONG __saveds __asm InitArchiveClassDataA(register __a0 Class *Cl, register __a1 struct TagItem *TagList)
{
	struct ArchiveClassData *acd, *ParentAcd;

	if (acd = AllocVec(sizeof(struct ArchiveClassData), MEMF_PUBLIC|MEMF_CLEAR)) {
		Cl->cl_UserData = (ULONG)acd;
		ParentAcd = (struct ArchiveClassData *)Cl->cl_Super->cl_UserData;
		acd->EntrySize = GetTagData(ACD_EntrySize, 0, TagList);
		acd->EntryOffset = ParentAcd->EntrySize + ParentAcd->EntryOffset;
		acd->HandleSize = GetTagData(ACD_HandleSize, 0, TagList);
		acd->HandleOffset = ParentAcd->HandleSize + ParentAcd->HandleOffset;
		return TRUE;
	}
	return FALSE;
}


void __saveds __asm FreeArchiveClassData(register __a0 Class *Cl)
{
	FreeVec((APTR)Cl->cl_UserData);
}

