/*
 *  ArchiveEntry.c - Copyright  1994 by S.R. & P.C.
 *
 *  Created:    05 Aug 1994  11:30:16
 *  Modified:   04 Dec 1994  11:22:32
 *
 *  Make>> smake
 */

#include <clib.h>
#include <Archives/ArchiveClass.h>

#include "Archive.h"
#include "Debug.h"

#include "proto/ArchiveEntry.h"
#include "proto/ASyncStuff.h"
#include "proto/Alloc.h"
#include "proto/Misc.h"

extern struct ArchiveBase *ArchiveBase;

#include <Archives/ArchiveMethods.h>


LONG ReleaseAccess(Object *O, struct PrivateArchiveEntry *ArchiveEntry, LONG Access)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);

	DBG_Trace();
	if (ArchiveEntry->ae_AccessCnt && ArchiveEntry->ae_Access == Access) {
		if (--ArchiveEntry->ae_AccessCnt == 0)
			ArchiveEntry->ae_Access = 0;
		Data->NumLocks--;
		return 0;
	}
	else
		return ERROR_INVALID_LOCK;
}


LONG GetAccess(Object *O, struct PrivateArchiveEntry *ArchiveEntry, LONG Access)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);

//TODO ERROR_IS_SOFT_LINK
	DBG_Trace();
	if (Access == EXCLUSIVE_LOCK) {
		if (ArchiveEntry->ae_Access)
			return ERROR_OBJECT_IN_USE;
		ArchiveEntry->ae_Access = EXCLUSIVE_LOCK;
		ArchiveEntry->ae_AccessCnt = 1;
		Data->NumLocks++;
		return 0;
	}
	else {		// SHARED_LOCK
		if (ArchiveEntry->ae_Access == EXCLUSIVE_LOCK)
			return ERROR_OBJECT_IN_USE;
		ArchiveEntry->ae_Access = SHARED_LOCK;
		ArchiveEntry->ae_AccessCnt++;
		Data->NumLocks++;
		return 0;
	}
}


struct FileLock *CreateLock(Object *O, struct PrivateArchiveEntry *ArchiveEntry, LONG Access, LONG *Error)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	struct FileLock *FileLock = NULL;

	if (!(*Error = GetAccess(O, ArchiveEntry, Access))) {
		if (FileLock = AllocArchiveObjectA(O, AO_Lock, NULL)) {
			FileLock->fl_Key = (LONG)ArchiveEntry;
			FileLock->fl_Access = Access;
			FileLock->fl_Task = Data->VolumeNode->dol_Task;
			FileLock->fl_Volume = MKBADDR(Data->VolumeNode);
		}
		else {
			ReleaseAccess(O, ArchiveEntry, Access);
			*Error = ERROR_NO_FREE_STORE;
		}
	}
	return FileLock;
}


void ArchiveEntryToFib(struct FileInfoBlock *FileInfoBlock, struct PrivateArchiveEntry *ArchiveEntry)
{
	FileInfoBlock->fib_DirEntryType = ArchiveEntry->ae_EntryType;
	FileInfoBlock->fib_EntryType = FileInfoBlock->fib_DirEntryType;		/* as recommended in the GuruBook */
	CToBStr(FileInfoBlock->fib_FileName, ArchiveEntry->ae_Name);
	if (ArchiveEntry->ae_Comment)
		CToBStr(FileInfoBlock->fib_Comment, ArchiveEntry->ae_Comment);
	else
		FileInfoBlock->fib_Comment[0] = 0;
	FileInfoBlock->fib_Protection = ArchiveEntry->ae_Protection;
	FileInfoBlock->fib_Size = ArchiveEntry->ae_Size;
	FileInfoBlock->fib_NumBlocks = ArchiveEntry->ae_NumBlocks;
	FileInfoBlock->fib_Date = ArchiveEntry->ae_Date;
	FileInfoBlock->fib_OwnerUID = ArchiveEntry->ae_OwnerUID;
	FileInfoBlock->fib_OwnerGID = ArchiveEntry->ae_OwnerGID;
}


void ScanArchive(Object *O)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);

	switch (Data->ScanState) {
	case SS_NotScanned:
		Data->ScanState = SS_ScanInProgress;
		DoAAHM_ScanArchive(O, TAG_END);
		Data->ScanState = SS_ScanDone;
		break;
	case SS_ScanInProgress:
		DoAAHM_WaitEndScan(O, TAG_END);
		break;
	}
}


static struct PrivateArchiveEntry *FindSubEntry(struct MinList *List, UBYTE *Name)
{
	struct PrivateArchiveEntry *ArchiveEntry;

	for (ArchiveEntry = GetHead(List); ArchiveEntry; ArchiveEntry = GetSucc(ArchiveEntry)) {
		if (!Stricmp(ArchiveEntry->ae_Name, Name))
			return ArchiveEntry;
	}
	return NULL;
}


struct PrivateArchiveEntry *FindArchiveEntry(Object *O, struct FileLock *FileLock, UBYTE *Path)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	struct PrivateArchiveEntry *ArchiveEntry;

	if (FileLock && FileLock->fl_Key)
		ArchiveEntry = (struct PrivateArchiveEntry *)FileLock->fl_Key;
	else
		ArchiveEntry = Data->RootEntry;
	if (Path) {
		char Part[40], *s;
		WORD Pos = 0;

		if (s = strchr(Path, ':')) {
			ArchiveEntry = Data->RootEntry;
			Path = s+1;
		}
		if (*Path) {	// Non empty path
			if (Data->ScanState != SS_ScanDone)
				ScanArchive(O);
			// Now parse the path.
			do {
				Pos = SplitName(Path, '/', Part, Pos, 39);
				DBG_Info1(KPrintF("Part: '%s'\n", Part));
				if (Part[0]) {
					if (ArchiveEntry->ae_EntryType > 0 && ArchiveEntry->ae_EntryType != ST_SOFTLINK) {
						if (!(ArchiveEntry = FindSubEntry(&ArchiveEntry->ae_SubEntries, Part)))
							break;
					}
					else {
						ArchiveEntry = NULL;
						break;
					}
				}
				else if (Pos != -1) {		// Always ignore trailing '/'
					if (ArchiveEntry->ae_Parent)
						ArchiveEntry = ArchiveEntry->ae_Parent;
				}
			} while (Pos != -1);
		}
	}
	return ArchiveEntry;
}


/*
 *	Scan...
 */

//TODO  gerer (et reporter) les erreurs


static struct PrivateArchiveEntry *CreateDirEntry(Object *O, UBYTE *Name)
{
	struct PrivateArchiveEntry *ArchiveEntry;

	if (ArchiveEntry = AllocArchiveObjectA(O, AO_Entry, NULL)) {
		strcpy(ArchiveEntry->ae_Name, Name);
		DateStamp(&ArchiveEntry->ae_Date);
		ArchiveEntry->ae_EntryType = ST_USERDIR;
	}
	return ArchiveEntry;
}


static struct PrivateArchiveEntry *GetParentFromDirPath(Object *O, UBYTE *Path)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	struct PrivateArchiveEntry *ArchiveEntry, *Parent;
	char Part[40], *s;
	WORD Pos = 0;

	ArchiveEntry = Data->RootEntry;		// Start from root
	if (s = strchr(Path, ':'))
		Path = s+1;

	if (*Path) {	// Non empty path
		do {
			Pos = SplitName(Path, '/', Part, Pos, 39);
			DBG_Info1(KPrintF("Part: '%s'\n", Part));
			Parent = ArchiveEntry;
			if (Part[0]) {
				if (ArchiveEntry = FindSubEntry(&Parent->ae_SubEntries, Part)) {
					if (ArchiveEntry->ae_EntryType != ST_USERDIR) {
						ArchiveEntry = NULL;
						break;
					}
				}
				else {
					if (!(ArchiveEntry = CreateDirEntry(O, Part)))
						break;
					AddHead((struct List *)&Parent->ae_SubEntries, (struct Node *)ArchiveEntry);
				}
			}
		} while (Pos != -1);
	}
	return ArchiveEntry;
}


static BOOL AddArchiveEntry(struct PrivateArchiveEntry *Parent, struct PrivateArchiveEntry *ArchiveEntry)
{
	// Put directories at head of the list for faster path resolution on later accesses
	if (ArchiveEntry->ae_EntryType == ST_USERDIR)
		AddHead((struct List *)&Parent->ae_SubEntries, (struct Node *)ArchiveEntry);
	else
		AddTail((struct List *)&Parent->ae_SubEntries, (struct Node *)ArchiveEntry);
	ArchiveEntry->ae_Parent = Parent;
	return TRUE;
}


ULONG aahm_ScanArchive(Class *Cl, Object *O, struct opTagList *M)
{
	struct aahpReadHeader ReadHeaderData;
	struct PrivateArchiveEntry *ArchiveEntry, *Parent;

	DBG(SimpleReq("Scan..."));
	memset(&ReadHeaderData, 0, sizeof(ReadHeaderData));
	DoAAHM_ReadFirstHeader(O, &ReadHeaderData, TAG_END);
	while (!ReadHeaderData.Error) {
		if (ArchiveEntry = (struct PrivateArchiveEntry *)ReadHeaderData.Entry) {
			DBG_Info1(KPrintF("Dir: %s  Entry: %s  Type: %ld  Size: %ld\n", 
					 ReadHeaderData.DirPath, ArchiveEntry->ae_Name, ArchiveEntry->ae_EntryType, ArchiveEntry->ae_Size));
			if ((Parent = ArchiveEntry->ae_Parent) || (Parent = GetParentFromDirPath(O, ReadHeaderData.DirPath))) {
				if (!AddArchiveEntry(Parent, ArchiveEntry))
					break;
			}
			else
				break;
			DoAAHM_ReadNextHeader(O, &ReadHeaderData, TAG_END);
		}
	}
	DoAAHM_ReadHeaderEnd(O, &ReadHeaderData, TAG_END);
	return 0;
}

