/*
 *  ASyncStuff.c - Copyright  1994 by S.R. & P.C.
 *
 *  Created:    21 Nov 1994  23:42:45
 *  Modified:   02 Dec 1994  00:41:02
 *
 *  Make>> smake
 */

#include <clib.h>
#include <archives/ArchiveClass.h>

#include "Archive.h"
#include "Debug.h"

#include "proto/ASyncStuff.h"
#include "proto/Handler.h"
#include "proto/Alloc.h"

extern struct ArchiveBase *ArchiveBase;

#include <archives/ArchiveMethods.h>


struct SlaveProcMessage {
	struct ASyncRequest spm_Request;
	UWORD spm_Type;
};

#define SPM_ASyncMethod		0
#define SPM_Finish			1


struct ASyncMethodRequest {
	struct SlaveProcMessage asmr_Request;
	Object *asmr_Object;
	Msg *asmr_Msg;
	ULONG asmr_RC;
};


ULONG __asm __saveds DoASyncMethodA(register __a0 Object *O, register __a1 Msg *M)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	struct ASyncMethodRequest *Req;
	ULONG RC = 0;

	if ((struct Process *)SysBase->ThisTask == Data->HandlerProcess) {
		if (Req = AllocArchiveObject(O, AO_ASyncRequest, AAO_ObjectSize, sizeof(struct ASyncMethodRequest), TAG_END)) {
			Req->spm_Type = SPM_ASyncMethod;
			Req->asmr_Object = O;
			Req->asmr_Msg = M;
			PutMsg(Data->SlavePort, Req);
			ASyncWaitA(O, Req, NULL);
			RC = Req->asmr_RC;
			FreeArchiveObject(O, AO_ASyncRequest, Req);
		}
		else
			RC = -2;		//TODO faire qqchose pour ce genre d'echec
	}
	else
		RC = DoMethodA(O, M);
	return RC;
}


void __asm __saveds ASyncWaitA(register __a0 Object *O, register __a1 struct ASyncRequest *ASyncRequest, register __a2 struct TagItem *TagList)
{
	void __asm ASyncWaitStub(register __a0 struct ASyncRequest *Req, register __a2 Object *O);
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);

	Data->ASyncRequestCnt++;
	ASyncWaitStub(ASyncRequest, O);
	Data->ASyncRequestCnt--;
}




struct SlaveProcArg {
	struct Message Message;
	Object *ArchiveObject;
	BOOL Success;			// Initialisation acknoledge
};


static void __saveds SlaveProcEntry(void)
{
	struct ArchiveData *Data;
	struct SlaveProcArg *SlaveProcArg;
	struct SlaveProcMessage *EndReq;
	struct Process *ThisProcess;
	Object *ArchiveObject;
	BOOL End = FALSE;

	DBG_Trace();
	DBG_Req();

	ThisProcess = (struct Process *)SysBase->ThisTask;
	WaitPort(&ThisProcess->pr_MsgPort);
	SlaveProcArg = (struct SlaveProcArg *)GetMsg(&ThisProcess->pr_MsgPort);
	ArchiveObject = SlaveProcArg->ArchiveObject;
	Data = INST_DATA(ArchiveBase->Class, ArchiveObject);

	EndReq = (struct SlaveProcMessage *)SlaveProcArg;	// In case of failure. reply after Forbid()
	if (Data->SlavePort = CreateMsgPort()) {
		SlaveProcArg->Success = TRUE;
		ReplyMsg(SlaveProcArg);

		while (!End) {
			struct SlaveProcMessage *Req;

			WaitPort(Data->SlavePort);
			while (Req = (struct SlaveProcMessage *)GetMsg(Data->SlavePort)) {
				switch (Req->spm_Type) {
				case SPM_ASyncMethod:
					{
						struct ASyncMethodRequest *mr = (struct ASyncMethodRequest *)Req;

						mr->asmr_RC = DoMethodA(mr->asmr_Object, mr->asmr_Msg);
						ReplyMsg(Req);
					}
					break;
				case SPM_Finish:
					EndReq = Req;	// Store it in case there would be other pending requests
					End = TRUE;
					break;
				}
			}
		}
	}
	Forbid();
	if (Data->SlavePort) {
		struct SlaveProcMessage *Req;

		while (Req = (struct SlaveProcMessage *)GetMsg(Data->SlavePort))
			ReplyMsg(Req);
		DeleteMsgPort(Data->SlavePort);
		Data->SlavePort = NULL;
	}
	ReplyMsg(EndReq);	// Reply the 'Finish' request (or startup msg if inits failed)
}



BOOL CreateSlaveProcess(Object *O)
{
	struct ArchiveData *const Data = INST_DATA(ArchiveBase->Class, O);
	struct Process *P;
	BOOL Ok = FALSE;

	P = CreateNewProcTags(NP_Entry, SlaveProcEntry,
						  NP_StackSize, STACKSIZE,
						  NP_Name, Data->SlaveProcName,
						  NP_Priority, 0,
						  NP_Input, Open("NIL:", MODE_OLDFILE),
						  NP_Output, Open("NIL:", MODE_NEWFILE),
						  NP_CurrentDir, NULL,
//						  NP_HomeDir, NULL,
						  NP_WindowPtr, NULL,
						  TAG_END);
	if (P) {
		__aligned struct SlaveProcArg SlaveProcArg;

		memset(&SlaveProcArg, 0, sizeof(struct SlaveProcArg));
		SlaveProcArg.Message.mn_ReplyPort = Data->ASyncReplyPort;
		SlaveProcArg.Message.mn_Length = sizeof(struct SlaveProcArg);
		SlaveProcArg.Message.mn_Node.ln_Type = NT_MESSAGE;
		SlaveProcArg.ArchiveObject = O;
		PutMsg(&P->pr_MsgPort, &SlaveProcArg);
		WaitPort(Data->ASyncReplyPort);
		GetMsg(Data->ASyncReplyPort);
		Ok = SlaveProcArg.Success;
	}
	return Ok;
}


void DeleteSlaveProcess(Object *O)
{
	struct ArchiveData *const Data = INST_DATA(ArchiveBase->Class, O);

	if (Data->SlavePort) {
		struct SlaveProcMessage EndMessage;

		memset(&EndMessage, 0, sizeof(EndMessage));
		EndMessage.mn_ReplyPort = Data->ASyncReplyPort;
		EndMessage.mn_Length = sizeof(EndMessage);
		EndMessage.mn_Node.ln_Type = NT_MESSAGE;
		EndMessage.spm_Type = SPM_Finish;
		PutMsg(Data->SlavePort, &EndMessage);
		WaitPort(Data->ASyncReplyPort);
		GetMsg(Data->ASyncReplyPort);
	}
}

