/*
 *  Handler.c - Copyright  1994 by S.R. & P.C.
 *
 *  Created:    23 Jan 1994
 *  Modified:   07 Dec 1994  12:42:39
 *
 *  Make>> smake
 */

#include <clib.h>
#include <archives/ArchiveClass.h>

#include "Archive.h"
#include "Debug.h"

#include "proto/Handler.h"
#include "proto/ASyncStuff.h"
#include "proto/ArchiveEntry.h"
#include "proto/Archive.h"
#include "proto/Alloc.h"
#include "proto/Misc.h"

extern struct ArchiveBase *ArchiveBase;

#include <archives/ArchiveMethods.h>


#define MOUNT_DOSLISTFLAGS		(LDF_WRITE|LDF_DEVICES|LDF_VOLUMES)

struct HandlerArg {
	struct Message Message;
	// Args TO the handler proc
	Object *ArchiveObject;

	// Results FROM the handler proc
	BPTR ArchiveRootLock;
};


static void __saveds HandlerEntry(void)
{
	struct ArchiveData *Data;
	struct HandlerArg *HandlerArg;
	struct Process *ThisProcess;
	struct DosList *DeviceNode;
	struct DosList *VolumeNode;
	Object *ArchiveObject;
	UBYTE *DOSDevBaseName, DOSDevName[40], *VolumeName;
	UBYTE *OldHandlerProcName;
	BOOL Success = FALSE;

	DBG_Trace();
	DBG_Req();
	ThisProcess = (struct Process *)SysBase->ThisTask;
	WaitPort(&ThisProcess->pr_MsgPort);
	HandlerArg = (struct HandlerArg *)GetMsg(&ThisProcess->pr_MsgPort);
	ArchiveObject = HandlerArg->ArchiveObject;
	Data = INST_DATA(ArchiveBase->Class, ArchiveObject);

	OpenLibrary("archive.library", 0);		// Open ourself to prevent unload. Handler will CloseLibrary before ending.

	// Get device & volume names...
	GetAttr(AA_DeviceBaseName, ArchiveObject, (ULONG *)&DOSDevBaseName);
	GetAttr(AA_VolumeName,     ArchiveObject, (ULONG *)&VolumeName);

	{
		struct DosList *dl;
		long i;

		dl = LockDosList(LDF_READ|LDF_DEVICES);
		i = 0;
		do {
			SPrintf(DOSDevName, "%s%ld", DOSDevBaseName, i++);
		} while (FindDosEntry(dl, DOSDevName, LDF_DEVICES));
		UnLockDosList(LDF_READ|LDF_DEVICES);
		SPrintf(Data->DeviceName, "%s:", DOSDevName);
		SPrintf(Data->SlaveProcName, "%s Slave", DOSDevName);
		DBG_Info(KPrintF("Dev: %s  Vol: %s\n", Data->DeviceName, VolumeName));
	}

	OldHandlerProcName = ThisProcess->tc_Node.ln_Name;
	ThisProcess->tc_Node.ln_Name = DOSDevName;

	if (DeviceNode = MakeDosEntry(DOSDevName, DLT_DEVICE)) {
		DeviceNode->dol_Task = &ThisProcess->pr_MsgPort;
		DBG_Req();
		if (VolumeNode = MakeDosEntry(VolumeName, DLT_VOLUME)) {
			DateStamp(&VolumeNode->dol_VolumeDate);
			VolumeNode->dol_Task = &ThisProcess->pr_MsgPort;

			// Store info object BEFORE calling AHM_Startup which needs it to create root lock.
			Data->DeviceNode = DeviceNode;
			Data->VolumeNode = VolumeNode;

			DBG_Req();
			if (HandlerArg->ArchiveRootLock = DoAHM_Startup(ArchiveObject, TAG_END)) {
				DBG_Req();

				// Process is not yet a handler. DosEntries not yet added to DosList.
				// So, LockDosList() is safe. (see autodocs for AddDosEntry())

				LockDosList(MOUNT_DOSLISTFLAGS);
				if (AddDosEntry(DeviceNode)) {
					/* Device is now mounted */
					DBG_Info1(KPrintF("%s, %s(), %l04d>Mounted\n", __FILE__, __FUNC__, __LINE__));
					if (AddDosEntry(VolumeNode)) {
						UnLockDosList(MOUNT_DOSLISTFLAGS);
						DBG_Info1(KPrintF("%s, %s(), %l04d>Volume added\n", __FILE__, __FUNC__, __LINE__));
						DBG_Req();
						do {
							struct ASyncRequest *DummyASyncReq;		// Not a request. Just for StackSwap mechanism

							DBG_Trace();
							DBG_Req();
							if (Data->Flags & ADF_UnMount) {
								DBG(SimpleReq("AttemptLockDosList() failed. go back to handler waitloop..."));
								SetTaskPri(ThisProcess, -1);	// Not sure Delay() is safe in a handler
							}
							if (DummyASyncReq = AllocArchiveObjectA(ArchiveObject, AO_ASyncRequest, NULL)) {
								// Handler will return Data->OriginalHandlerStack when living,
								// so that ASyncWaitStub can StackSwap().
								Data->OriginalHandlerStack = DummyASyncReq;
								
								if (!(Data->Flags & ADF_UnMount)) {		// First time
									ReplyMsg(HandlerArg);
									Success = TRUE;
								}
								ASyncWaitA(ArchiveObject, DummyASyncReq, NULL);	// Swap Stack & call HandlerWaitLoop()
								// Now free the dummy request. The request IS the original one.
								// But stack may not be if async mechanism has been used.
								FreeArchiveObject(ArchiveObject, AO_ASyncRequest, DummyASyncReq);
							}
							DBG_Trace();
							DBG_Req();
							Data->Flags |= ADF_UnMount;		// further calls to WaitLoop will not Wait()
						} while (!AttemptLockDosList(MOUNT_DOSLISTFLAGS));
						RemDosEntry(VolumeNode);
					}
					else
						DBG_Info1(KPrintF("%s, %s(), %l04d>Failed to add volume\n", __FILE__, __FUNC__, __LINE__));
					DBG_Req();
					RemDosEntry(DeviceNode);
				}
				else
					DBG_Info1(KPrintF("%s, %s(), %l04d>Failed to add device\n", __FILE__, __FUNC__, __LINE__));
				DBG_Req();
				UnLockDosList(MOUNT_DOSLISTFLAGS);
			}
			DBG_Req();
			FreeDosEntry(VolumeNode);
			DBG_Req();
		}
		FreeDosEntry(DeviceNode);
		DBG_Req();
	}

	DoAHM_Finish(ArchiveObject, TAG_END);

	if (Success) {
		DBG(SimpleReq("About to DisposeArchive()..."));
		DisposeArchive(ArchiveObject);		// Dispose object only if mount was successfull. else, done by caller.
	}
	DBG(SimpleReq("Dispose done, Forbid() and quit..."));

	Forbid();					// prevent to be unloaded (code is in archive.library) before really dead
	ThisProcess->tc_Node.ln_Name = OldHandlerProcName;
	if (!Success) {
		HandlerArg->ArchiveRootLock = NULL;
		ReplyMsg(HandlerArg);
	}
	CloseLibrary(ArchiveBase);
}


BPTR am_Mount(Class *Cl, Object *O, struct opTagList *M)
{
	struct ArchiveData *const Data = INST_DATA(Cl, O);
	struct HandlerArg *HandlerArg;
	BPTR RootLock = NULL;
	struct MsgPort *ReplyPort;

	DBG_Trace();
	DBG_Req();
	if (GetTagData(MA_AutoUnmount, FALSE, M->TagList))
		Data->Flags |= ADF_QuitHandler;

	if (HandlerArg = AllocVec(sizeof(struct HandlerArg), MEMF_PUBLIC|MEMF_CLEAR)) {
		if (ReplyPort = CreateMsgPort()) {
			HandlerArg->Message.mn_ReplyPort = ReplyPort;
			HandlerArg->Message.mn_Length = sizeof(struct HandlerArg);
			HandlerArg->Message.mn_Node.ln_Type = NT_MESSAGE;
			HandlerArg->ArchiveObject = O;

			Data->HandlerProcess = CreateNewProcTags(NP_Entry, HandlerEntry,
													 NP_StackSize, STACKSIZE,
													 NP_Priority, 10,
													 NP_Input, NULL,
													 NP_Output, NULL,
													 NP_CloseInput, FALSE,
													 NP_CloseOutput, FALSE,
													 NP_Error, NULL,
													 NP_CloseError, FALSE,
													 NP_CurrentDir, NULL,
													 NP_HomeDir, NULL,
													 NP_ConsoleTask, NULL,
													 NP_WindowPtr, NULL,
													 TAG_END);

			DBG_Req();

			if (Data->HandlerProcess) {
				DBG_Req();
				PutMsg(&Data->HandlerProcess->pr_MsgPort, HandlerArg);
				WaitPort(ReplyPort);
				GetMsg(ReplyPort);
				RootLock = HandlerArg->ArchiveRootLock;
				DBG_Req();
			}
			DeleteMsgPort(ReplyPort);
		}
		FreeVec(HandlerArg);
	} 
	DBG_Req();
	return RootLock;
}


BPTR ahm_Startup(Class *Cl, Object *O, struct opTagList *M)
{
	struct ArchiveData *const Data = INST_DATA(Cl, O);
	LONG Error;

	DBG_Trace();
	// Create port in handler context (not in OM_NEW)
	if (Data->ASyncReplyPort = CreateMsgPort()) {
		if (CreateSlaveProcess(O)) {
			if (Data->RootEntry = AllocArchiveObjectA(O, AO_Entry, NULL)) {
				BToCStr(Data->RootEntry->ae_Name, Data->VolumeNode->dol_Name);
				Data->RootEntry->ae_EntryType = ST_ROOT;
				Data->RootEntry->ae_Date = Data->VolumeNode->dol_VolumeDate;
				return MKBADDR(CreateLock(O, Data->RootEntry, ACCESS_READ, &Error));
			}
			DeleteSlaveProcess(O);
		}
		// Something failed. Make sure reply port is deleted by the same process that created it
		DeleteMsgPort(Data->ASyncReplyPort);
		Data->ASyncReplyPort = NULL;
	}
	return NULL;
}


void ahm_Finish(Class *Cl, Object *O, struct opTagList *M)
{
	struct ArchiveData *const Data = INST_DATA(Cl, O);

	DBG_Trace();
	DBG_Req();
	DeleteSlaveProcess(O);
	if (Data->ASyncReplyPort)
		DeleteMsgPort(Data->ASyncReplyPort);
}


static struct DosPacket *GetPacket(struct MsgPort *MsgPort)
{
	struct Message *Msg;

	if (Msg = GetMsg(MsgPort)) {
		#ifdef DEBUG
		if (!Msg->mn_Node.ln_Name)
			SimpleReq("%s, %s(), %l04d>E:Not a Dos packet", __FILE__, __FUNC__, __LINE__);
		#endif
		return (struct DosPacket *)Msg->mn_Node.ln_Name;
	}
	return NULL;
}


struct ASyncRequest * __asm __saveds HandlerWaitLoop(register __a0 Object *O)
{
	struct ArchiveData *const Data = INST_DATA(ArchiveBase->Class, O);
	struct DosPacket *DosPacket = NULL;
	struct ASyncRequest *ASyncReq;
	struct MsgPort *HandlerPort;
	ULONG SignalMask;

	HandlerPort = &((struct Process *)SysBase->ThisTask)->pr_MsgPort;
	SignalMask = (1L << HandlerPort->mp_SigBit) | (1L << Data->ASyncReplyPort->mp_SigBit);
	DBG_Req();
	for (;;) {
		for (;;) {
			if (Data->ASyncReplyPort && (ASyncReq = (struct ASyncRequest *) GetMsg(Data->ASyncReplyPort))) {
				DBG_Info1(KPrintF("Got ASyncRequest Reply\n"));
				return ASyncReq;
			}
			else if (DosPacket = GetPacket(HandlerPort)) {
				DBG_Info1(KPrintF("Packet: %04ld  Arg1: %ld  Arg2: %ld  Arg3: %ld\n", DosPacket->dp_Type,
								   DosPacket->dp_Arg1, DosPacket->dp_Arg2, DosPacket->dp_Arg3));
				DosPacket->dp_Res1 = DOSFALSE;		/* force to put the right Result. */
				DosPacket->dp_Res2 = ERROR_ACTION_NOT_KNOWN;
				DoAHM_DoPacket(O, DosPacket, TAG_END);
				ReplyPkt(DosPacket, DosPacket->dp_Res1, DosPacket->dp_Res2);
				// ASyncRequestCnt is always at least 1 since 
				if ((Data->Flags & ADF_QuitHandler) && Data->ASyncRequestCnt == 1 && Data->NumLocks == 0) {
					//DBG(SimpleReq("About to quit handler..."));
					return Data->OriginalHandlerStack;
				}
			}
			else
				break;
		}
		// Do not Wait() if ADF_UnMount. AttempLockDosList() failed and then tried to process eventual new packets
		if ((Data->Flags & ADF_UnMount) && Data->ASyncRequestCnt == 1 && Data->NumLocks == 0) {
			DBG(SimpleReq("ADF_UnMount flag detected. Don't go to Wait() and live..."));
			break;
		}
		Wait(SignalMask);
	}
	return Data->OriginalHandlerStack;
}


BPTR __saveds __asm MountArchiveA(register __a0 UBYTE *FileName, register __a1 struct TagItem *TagList)
{
	BPTR ArcVolumeLock = 0;
	Object *O;

	/*	To prevent failure to add device to device list because of dupplicate name,
	 *	protect mount section with a semaphore. This is not perfect but better than nothing!
	 *	This semaphore also protect the list of mounted archives, list used to prevent
	 *	multiple mount of the same archive. It is not a limitation, but a need.
	 *	When archive write will be implemented, volume will need ACCESS_WRITE (exclusive) right on archive.
	 */
	ObtainSemaphore(&ArchiveBase->MountSemaphore);

//TODO
	// Chercher si 'FileLock' n'est pas deja une archive mountee...
	// si non: mettre dans la liste tout de suite.	(file lock seul pour l'instant)

	if (O = NewArchiveA(FileName, TagList)) {
		if (ArcVolumeLock = (BPTR)DoMethod(O, AM_Mount, TagList)) {
			UBYTE *DeviceName, *dn;
			
			if (DeviceName = (UBYTE *)GetTagData(MA_DeviceName, NULL, TagList)) {
				GetAttr(AA_DeviceName, O, (ULONG *)&dn);
				strcpy(DeviceName, dn);
			}
		}
		else {
			DBG(SimpleReq("Mount Failed. Dispose()..."));
			DisposeArchive(O);	// DO NOT Dispose() in Mount succeded. Will be done by handler at UnMount time
		}
	}

	// si  fail, remove from list
	// si Ok, mettre a jour l'element de la liste. (mettre l'objet archive dedans)

	ReleaseSemaphore(&ArchiveBase->MountSemaphore);
	return ArcVolumeLock;
}


