/*
 *  LibInit.c - Copyright  1994 by S.R. & P.C.
 *
 *  Created:    21 Jul 1994  21:55:22
 *  Modified:   07 Dec 1994  12:31:31
 *
 *  Make>> smake
 */

#include <clib.h>

#include "Archive.h"
#include "Archive_rev.h"

#include "proto/Archive.h"
#include "proto/Handler.h"
#include "proto/ASyncStuff.h"
#include "proto/Alloc.h"
#include "proto/Misc.h"

typedef void *(*PVF)();

/* library initialization table, used for AUTOINIT libraries                */
struct InitTable {
	ULONG it_DataSize;			/* library data space size         */
	PVF *it_FuncTable;			/* table of entry points           */
	APTR it_DataInit;			/* table of data initializers      */
	void (*it_InitFunc) (void);	/* initialization function to run  */
};

BPTR __saveds __asm LibExpunge(register __a6 struct ArchiveBase *LibBase);
struct ArchiveBase *__saveds __asm LibInit(register __a0 BPTR seglist, register __d0 struct ArchiveBase *LibBase);
struct ArchiveBase *__saveds __asm LibOpen(register __a6 struct ArchiveBase *LibBase);
BPTR __saveds __asm LibClose(register __a6 struct ArchiveBase *LibBase);

far extern struct Resident LibRomTag;	/* Come from ASM */

struct ExecBase *SysBase;
struct DosLibrary *DOSBase;
struct Library *UtilityBase;
struct Library *WhatIsBase;
struct IntuitionBase *IntuitionBase;
struct ArchiveBase *ArchiveBase;

char LibID[] = VSTRING;		// Do not use VERSTAG. No '$VER:' in rom tags!
char LibName[] = "archive.library";

char *ReqTitle = LibName;	// For SimpleReq()

/* NEVER forget the far or SASC put the offset from data segment, not the Ptr */
far PVF Lib_FuncTab[] = {
	(PVF) LibOpen,				/* standard open        */
	(PVF) LibClose,				/* standard close       */
	(PVF) LibExpunge,			/* standard expunge     */
	0,							/* futur Commodore expansion */

/* real lib functions */
	(PVF) NewArchiveA,
	(PVF) DisposeArchive,
	(PVF) MountArchiveA,
	(PVF) DoASyncMethodA,
	(PVF) ASyncWaitA,
	(PVF) AllocArchiveObjectA,
	(PVF) FreeArchiveObject,
	(PVF) InitArchiveClassDataA,
	(PVF) FreeArchiveClassData,
	(PVF) AddArchiveTypeA,
	(PVF) - 1L
};


far struct InitTable LibInitTab = {
	sizeof(struct ArchiveBase),
	Lib_FuncTab,
	NULL,						/* will initialize my own data */
	(void (*)()) LibInit,
};


struct ArchiveBase *__saveds __asm LibInit(register __a0 BPTR seglist, register __d0 struct ArchiveBase *LibBase)
{
	SysBase = *((struct ExecBase **) 4L);

	LibBase->SegList = seglist;

	/* init. library structure (since I don't do automatic data init.) */
	LibBase->Lib.lib_Node.ln_Type = NT_LIBRARY;
	LibBase->Lib.lib_Node.ln_Name = LibName;
	LibBase->Lib.lib_Flags = LIBF_SUMUSED | LIBF_CHANGED;
	LibBase->Lib.lib_Version = VERSION;
	LibBase->Lib.lib_Revision = REVISION;
	LibBase->Lib.lib_IdString = (APTR)LibID;
	InitSemaphore(&LibBase->MountSemaphore);
	NewMinList(&LibBase->ArchiveTypes);
	return (LibBase);
}


struct ArchiveBase *__saveds __asm LibOpen(register __a6 struct ArchiveBase *LibBase)
{
	ArchiveBase = LibBase;
	if (LibBase->Lib.lib_OpenCnt == 0) {
		if (!(DOSBase = (struct DosLibrary *) OpenLibrary("dos.library", 37L)))
			return NULL;
		UtilityBase = DOSBase->dl_UtilityBase;
		IntuitionBase = (struct IntuitionBase *)OpenLibrary("intuition.library", 37L);
		if (!WhatIsBase) {
			// Keep whatis.library open and types loaded until expunge to prevent parsing
			// again all types (in DEVS:Archives) after each time OpenCnt fall down to 0.
			if (!(WhatIsBase = OpenLibrary("whatis.library", 3L))) {
				SimpleReq("Requires whatis.library v3");
				return NULL;
			}
			LoadArchiveTypes();
		}
		if (!(LibBase->Class = InitCl()))
			return NULL;
	}
	/* mark us as having another customer */
	LibBase->Lib.lib_OpenCnt++;
	/* clear delayed expunges (standard procedure)                */
	LibBase->Lib.lib_Flags &= ~LIBF_DELEXP;
	return (LibBase);
}


BPTR __saveds __asm LibClose(register __a6 struct ArchiveBase * LibBase)
{
	BPTR retval = NULL;

	if (--LibBase->Lib.lib_OpenCnt == 0) {
		FreeCl(LibBase->Class);
		LibBase->Class = NULL;	/* security */
		CloseLibrary(IntuitionBase);
		CloseLibrary(DOSBase);
		if (LibBase->Lib.lib_Flags & LIBF_DELEXP) {
			/* no more people have me open, and I have a delayed expunge pending */
			retval = LibExpunge(LibBase);	/* return segment list        */
		}
	}
	return (retval);
}


BPTR __saveds __asm LibExpunge(register __a6 struct ArchiveBase * LibBase)
{
	BPTR SegList = NULL;
	LONG LibSize;

	LibBase->Lib.lib_Flags |= LIBF_DELEXP;
	if (LibBase->Lib.lib_OpenCnt == 0) {
		FreeArchiveTypes();
		CloseLibrary(WhatIsBase);
		Remove((struct Node *) LibBase);
		/* really expunge: remove LibBase and freemem        */
		SegList = LibBase->SegList;
		LibSize = LibBase->Lib.lib_NegSize + LibBase->Lib.lib_PosSize;
		FreeMem((char *) LibBase - LibBase->Lib.lib_NegSize, (LONG) LibSize);
	}
	else
		LibBase->Lib.lib_Flags |= LIBF_DELEXP;

	/* return NULL or real seglist                                */
	return (SegList);
}

