/*
 *  Packet.c - Copyright  1994 by S.R. & P.C.
 *
 *  Created:    23 Jan 1994
 *  Modified:   05 Dec 1994  17:38:14
 *
 *  Make>> smake
 */

#include <clib.h>
#include <Archives/ArchiveClass.h>

#include "Archive.h"
#include "Debug.h"

#include "proto/Packet.h"
#include "proto/ASyncStuff.h"
#include "proto/ArchiveEntry.h"
#include "proto/Alloc.h"
#include "proto/Misc.h"

extern struct ArchiveBase *ArchiveBase;

#include <Archives/ArchiveMethods.h>


#define BLOCK_SIZE	512


#ifdef DEBUG

static void PacketName(LONG Type)
{
	switch (Type) {
	case ACTION_DIE:
		KPrintF("DIE\n"); break;
	case ACTION_FINDUPDATE:
		KPrintF("FINDUPDATE\n"); break;
	case ACTION_FINDINPUT:
		KPrintF("FINDINPUT\n"); break;
	case ACTION_FINDOUTPUT:
		KPrintF("FINDOUTPUT\n"); break;
	case ACTION_READ:
		KPrintF("READ\n"); break;
	case ACTION_WRITE:
		KPrintF("WRITE\n"); break;
	case ACTION_END:
		KPrintF("CLOSE\n"); break;
	case ACTION_SEEK:
		KPrintF("SEEK\n"); break;
	case ACTION_EXAMINE_NEXT:
		KPrintF("EXAMINE NEXT\n"); break;
	case ACTION_EXAMINE_OBJECT:
		KPrintF("EXAMINE\n"); break;
	case ACTION_INFO:
		KPrintF("INFO\n"); break;
	case ACTION_DISK_INFO:
		KPrintF("DISK INFO\n"); break;
	case ACTION_PARENT:
		KPrintF("PARENTDIR\n"); break;
	case ACTION_DELETE_OBJECT:
		KPrintF("DELETE\n"); break;
	case ACTION_CREATE_DIR:
		KPrintF("CREATEDIR\n"); break;
	case ACTION_LOCATE_OBJECT:
		KPrintF("LOCK\n"); break;
	case ACTION_COPY_DIR:
		KPrintF("DUPLOCK\n"); break;
	case ACTION_FREE_LOCK:
		KPrintF("FREELOCK\n"); break;
	case ACTION_SET_PROTECT:
		KPrintF("SETPROTECT\n"); break;
	case ACTION_SET_COMMENT:
		KPrintF("SETCOMMENT\n"); break;
	case ACTION_RENAME_OBJECT:
		KPrintF("RENAME\n"); break;
	case ACTION_INHIBIT:
		KPrintF("INHIBIT\n"); break;
	case ACTION_RENAME_DISK:
		KPrintF("RENAME DISK\n"); break;
	case ACTION_MORE_CACHE:
		KPrintF("MORE CACHE\n"); break;
	case ACTION_FLUSH:
		KPrintF("FLUSH\n"); break;
	case ACTION_IS_FILESYSTEM:
		KPrintF("IS_FILESYSTEM\n"); break;
	case ACTION_SAME_LOCK:
		KPrintF("SAME_LOCK\n"); break;
	case ACTION_COPY_DIR_FH:
		KPrintF("COPY_DIR_FH\n"); break;
	case ACTION_PARENT_FH:
		KPrintF("PARENT_FH\n"); break;
	case ACTION_EXAMINE_FH:
		KPrintF("EXAMINE_FH\n"); break;
	case ACTION_FH_FROM_LOCK:
		KPrintF("FH_FROM_LOCK\n"); break;
	case ACTION_CURRENT_VOLUME:
		KPrintF("CURRENT_VOLUME\n"); break;
	case ACTION_READ_LINK:
		KPrintF("READ LINK\n"); break;
	case ACTION_MAKE_LINK:
		KPrintF("MAKE LINK\n"); break;
	case ACTION_EXAMINE_ALL:
		KPrintF("EXAMINE ALL\n"); break;
	default:
		KPrintF("------ UNKNOWN (%ld) ------\n", Type);
	}
}

#endif


static void InitInfoData(Object *O, struct InfoData *InfoData)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);

	InfoData->id_NumSoftErrors = 0;
	InfoData->id_UnitNumber = 0;
	InfoData->id_DiskState = ID_WRITE_PROTECTED;	/* Default setting */
	InfoData->id_NumBlocks = 1;		// Prevent zero divide in apps that does not check it.
	InfoData->id_NumBlocksUsed = InfoData->id_NumBlocks;
	InfoData->id_BytesPerBlock = BLOCK_SIZE;
	InfoData->id_DiskType = ID_DOS_DISK;
	InfoData->id_VolumeNode = MKBADDR(Data->VolumeNode);
	InfoData->id_InUse = 0;				// Meaningfull for console handlers only.
}


static ULONG do_Packet_DISK_INFO(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct InfoData *InfoData;

	InfoData = BADDR(DosPacket->dp_Arg1);
	InitInfoData(O, InfoData);
	GetAttr(AA_VolumeInfo, O, (ULONG *)InfoData);
	DosPacket->dp_Res1 = DOSTRUE;
	DosPacket->dp_Res2 = 0;
	return TRUE;
}


static ULONG do_Packet_INFO(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct InfoData *InfoData;

	InfoData = BADDR(DosPacket->dp_Arg2);
	InitInfoData(O, InfoData);
	GetAttr(AA_VolumeInfo, O, (ULONG *)InfoData);
	DosPacket->dp_Res1 = DOSTRUE;
	DosPacket->dp_Res2 = 0;
	return TRUE;
}


static ULONG do_Packet_FREE_LOCK(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveEntry *ArchiveEntry;
	struct FileLock *FileLock;

	/*
	 *	dp_Arg1: BPTR to struct FileLock
	 *
	 *	dp_Res1: DOSTRUE/DOSFALSE
	 *	dp_Res2: LONG error code (if dp_Res1 == DOSFALSE)
	 */
	DBG_Trace();
	/* ZERO/root lock, do nothing as recommended in GuruBook */
	/* but it is in the strict sense an error of the application */
	if (FileLock = BADDR(DosPacket->dp_Arg1)) {
		if (ArchiveEntry = (struct PrivateArchiveEntry *)FileLock->fl_Key)
			ReleaseAccess(O, ArchiveEntry, FileLock->fl_Access);
		FileLock->fl_Task = NULL;	// Just a hack! recommended in guru book though
		FreeArchiveObject(O, AO_Lock, FileLock);
	}
	DosPacket->dp_Res1 = DOSTRUE;
	DosPacket->dp_Res2 = 0;
	return 0;
}


static ULONG do_Packet_LOCATE_OBJECT(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveEntry *ArchiveEntry;
	struct FileLock *FileLock;
	UBYTE Path[256];

	/*
	 *	dp_Arg1: BPTR to struct FileLock of parent
	 *	dp_Arg2: BPTR to BSTR of name
	 *	dp_Arg3: LONG access mode
	 *
	 *	dp_Res1: BPTR to struc FileLock of object
	 *	dp_Res2: LONG error code (if dp_Res1 == 0)
	 */
	DBG_Trace();
	BToCStr(Path, DosPacket->dp_Arg2);
	if (ArchiveEntry = FindArchiveEntry(O, BADDR(DosPacket->dp_Arg1), Path)) {
		if (FileLock = CreateLock(O, ArchiveEntry, DosPacket->dp_Arg3, &DosPacket->dp_Res2))
			DosPacket->dp_Res1 = MKBADDR(FileLock);
	}
	else
		DosPacket->dp_Res2 = ERROR_OBJECT_NOT_FOUND;
	return 0;
}


static ULONG do_Packet_COPY_DIR(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	struct FileLock *FileLock;

	/*
	 *	dp_Arg1: BPTR to struct FileLock (original)
	 *
	 *	dp_Res1: BPTR to struct FileLock (copy)
	 *	dp_Res2: LONG error code (if dp_Res1 == 0)
	 */

	DBG_Trace();
	if (FileLock = BADDR(DosPacket->dp_Arg1)) {
		if (FileLock = CreateLock(O, (struct PrivateArchiveEntry *)FileLock->fl_Key, SHARED_LOCK, &DosPacket->dp_Res2))
			DosPacket->dp_Res1 = MKBADDR(FileLock);
	}
	else {
		if (FileLock = CreateLock(O, Data->RootEntry, SHARED_LOCK, &DosPacket->dp_Res2))
			DosPacket->dp_Res1 = MKBADDR(FileLock);
	}
	return 0;
}


static ULONG do_Packet_COPY_DIR_FH(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveHandle *ArchiveHandle;
	struct FileLock *FileLock;

	/*
	 *	dp_Arg1: FileHandle->fh_Arg1;
	 *
	 *	dp_Res1: BPTR to struct FileLock (copy)
	 *	dp_Res2: LONG error code (if dp_Res1 == 0)
	 */

	DBG_Trace();
	if (ArchiveHandle = (struct PrivateArchiveHandle *)DosPacket->dp_Arg1) {
		struct PrivateArchiveEntry *ArchiveEntry = (struct PrivateArchiveEntry *)ArchiveHandle->ah_Entry;

		if (FileLock = CreateLock(O, ArchiveEntry, ArchiveHandle->ah_Access, &DosPacket->dp_Res2))
			DosPacket->dp_Res1 = MKBADDR(FileLock);
	}
	else
		DosPacket->dp_Res2 = ERROR_INVALID_LOCK;
	return 0;
}


static ULONG do_Packet_PARENT(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveEntry *ArchiveEntry;
	struct FileLock *FileLock;

	/*
	 *	dp_Arg1: BPTR to struct FileLock to get parent of
	 *
	 *	dp_Res1: BPTR to struc FileLock (parent lock)
	 *	dp_Res2: LONG error code (if dp_Res1 == 0) with special case.
	 */
	DBG_Trace();
	DosPacket->dp_Res2 = 0;
	if (FileLock = BADDR(DosPacket->dp_Arg1)) {
		if (ArchiveEntry = (struct PrivateArchiveEntry *)FileLock->fl_Key) {
			if (ArchiveEntry->ae_Parent) {
				if (FileLock = CreateLock(O, ArchiveEntry->ae_Parent, SHARED_LOCK, &DosPacket->dp_Res2))
					DosPacket->dp_Res1 = MKBADDR(FileLock);
			}
		}
	}
	return 0;
}


static ULONG do_Packet_PARENT_FH(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveHandle *ArchiveHandle;
	struct PrivateArchiveEntry *ArchiveEntry;
	struct FileLock *FileLock;

	/*
	 *	dp_Arg1: FileHandle->fh_Arg1;
	 *
	 *	dp_Res1: BPTR to struc FileLock (parent lock)
	 *	dp_Res2: LONG error code (if dp_Res1 == 0) with special case.
	 */

	DBG_Trace();
	DosPacket->dp_Res2 = 0;
	if (ArchiveHandle = (struct PrivateArchiveHandle *)DosPacket->dp_Arg1) {
		ArchiveEntry = (struct PrivateArchiveEntry *)ArchiveHandle->ah_Entry;
		if (ArchiveEntry->ae_Parent) {
			if (FileLock = CreateLock(O, ArchiveEntry->ae_Parent, SHARED_LOCK, &DosPacket->dp_Res2))
				DosPacket->dp_Res1 = MKBADDR(FileLock);
		}
	}
	else
		DosPacket->dp_Res2 = ERROR_INVALID_LOCK;
	return 0;
}


static ULONG do_Packet_EXAMINE_FH(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveHandle *ArchiveHandle;
	struct FileInfoBlock *FileInfoBlock;

	/*
	 *	dp_Arg1: FileHandle->fh_Arg1;
	 *	dp_Arg2: BPTR to struct FileInfoBlock to fill in
	 *
	 *	dp_Res1: DOSTRUE/DOSFALSE
	 *	dp_Res2: LONG error code (if dp_Res1 == DOSFALSE)
	 */

	DBG_Trace();
	FileInfoBlock = BADDR(DosPacket->dp_Arg2);
	if (ArchiveHandle = (struct PrivateArchiveHandle *)DosPacket->dp_Arg1) {
		ArchiveEntryToFib(FileInfoBlock, (struct PrivateArchiveEntry *)ArchiveHandle->ah_Entry); 
		DosPacket->dp_Res1 = DOSTRUE;
	}
	else
		DosPacket->dp_Res2 = ERROR_INVALID_LOCK;
	return 0;
}


static ULONG do_Packet_EXAMINE_OBJECT(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	struct PrivateArchiveEntry *ArchiveEntry;
	struct FileInfoBlock *FileInfoBlock;
	struct FileLock *FileLock;

	/*
	 *	dp_Arg1: BPTR to struct FileLock
	 *	dp_Arg2: BPTR to struct FileInfoBlock to fill in
	 *
	 *	dp_Res1: DOSTRUE/DOSFALSE
	 *	dp_Res2: LONG error code (if dp_Res1 == DOSFALSE)
	 */

	DBG_Trace();
	FileLock = BADDR(DosPacket->dp_Arg1);
	FileInfoBlock = BADDR(DosPacket->dp_Arg2);
	if (!FileLock || !(ArchiveEntry = (struct PrivateArchiveEntry *)FileLock->fl_Key))
		ArchiveEntry = Data->RootEntry;
	ArchiveEntryToFib(FileInfoBlock, ArchiveEntry); 
	// Prepare for ExNext()
	FileInfoBlock->fib_DiskKey = (LONG)GetHead(&ArchiveEntry->ae_SubEntries);
	DosPacket->dp_Res1 = DOSTRUE;
	return 0;
}


static ULONG do_Packet_EXAMINE_NEXT(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct ArchiveData *Data = INST_DATA(ArchiveBase->Class, O);
	struct PrivateArchiveEntry *ArchiveEntry;
	struct FileInfoBlock *FileInfoBlock;

	/*
	 *	dp_Arg1: BPTR to struct FileLock
	 *	dp_Arg2: BPTR to struct FileInfoBlock to fill in
	 *
	 *	dp_Res1: LONG DOSTRUE/DOSFALSE
	 *	dp_Res2: LONG error code (if dp_Res1 == DOSFALSE)
	 */
	DBG_Trace();
	FileInfoBlock = BADDR(DosPacket->dp_Arg2);

	if (Data->ScanState != SS_ScanDone) {
		ScanArchive(O);
		// If scan was not done, fib refer to root directory.
		// Re-init fib_DiskKey to what it should have been if archive was scanned before Examine()
		FileInfoBlock->fib_DiskKey = (LONG)GetHead(&Data->RootEntry->ae_SubEntries);
	}
	if (ArchiveEntry = (struct PrivateArchiveEntry *)FileInfoBlock->fib_DiskKey) {
		ArchiveEntryToFib(FileInfoBlock, ArchiveEntry); 
		FileInfoBlock->fib_DiskKey = (LONG)GetSucc(ArchiveEntry);
		DosPacket->dp_Res1 = DOSTRUE;
		DosPacket->dp_Res2 = 0;
	}
	else
		DosPacket->dp_Res2 = ERROR_NO_MORE_ENTRIES;
	return 0;
}


static ULONG do_Packet_FH_FROM_LOCK(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveEntry *ArchiveEntry;
	struct PrivateArchiveHandle *ArchiveHandle;
	struct FileHandle *FileHandle;
	struct FileLock *FileLock;

	/*
	 *	dp_Arg1: BPTR to struct FileHandle
	 *	dp_Arg2: BPTR to struct FileLock
	 *
	 *	dp_Res1: LONG DOSTRUE/DOSFALSE
	 *	dp_Res2: LONG error code (if dp_Res1 == DOSFALSE)
	 */

	DBG_Trace();
	FileHandle = BADDR(DosPacket->dp_Arg1);
	FileLock = BADDR(DosPacket->dp_Arg2);
	if (ArchiveEntry = (struct PrivateArchiveEntry *)FileLock->fl_Key) {
		if (ArchiveEntry->ae_EntryType == ST_FILE || ArchiveEntry->ae_EntryType == ST_LINKFILE) {
			if (FileLock->fl_Access == ACCESS_READ) {
				if (ArchiveHandle = AllocArchiveObjectA(O, AO_Handle, NULL)) {
					ArchiveHandle->ah_FileHandle = FileHandle;
					ArchiveHandle->ah_Entry = ArchiveEntry;
					ArchiveHandle->ah_Mode = MODE_OLDFILE;
					ArchiveHandle->ah_Access = SHARED_LOCK;
					FileHandle->fh_Arg1 = (LONG)ArchiveHandle;
					DosPacket->dp_Res2 = 0;
					DoAAHM_Open(O, ArchiveHandle, &DosPacket->dp_Res2, TAG_END);
					if (DosPacket->dp_Res2 == 0) {
						DosPacket->dp_Res1 = DOSTRUE;
						FreeArchiveObject(O, AO_Lock, FileLock);	// Lock is absorbed
						return 0;
					}
					FreeArchiveObject(O, AO_Entry, ArchiveHandle);
				}
				else
					DosPacket->dp_Res2 = ERROR_NO_FREE_STORE;
			}
			else
				DosPacket->dp_Res2 = ERROR_DISK_WRITE_PROTECTED;
		}
		else
			DosPacket->dp_Res2 = ERROR_OBJECT_WRONG_TYPE;
	}
	else
		DosPacket->dp_Res2 = ERROR_INVALID_LOCK;
	return 0;
}


static ULONG do_Packet_FINDINPUT(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveEntry *ArchiveEntry;
	struct PrivateArchiveHandle *ArchiveHandle;
	struct FileHandle *FileHandle;
	UBYTE Path[256];

	/*
	 *	dp_Arg1: BPTR to struct FileHandle
	 *	dp_Arg2: BPTR to struct FileLock
	 *	dp_Arg3: BPTR to BSTR Name
	 *
	 *	dp_Res1: LONG DOSTRUE/DOSFALSE
	 *	dp_Res2: LONG error code (if dp_Res1 == DOSFALSE)
	 */

	DBG_Trace();
	FileHandle = BADDR(DosPacket->dp_Arg1);
	BToCStr(Path, DosPacket->dp_Arg3);
	if (ArchiveEntry = FindArchiveEntry(O, BADDR(DosPacket->dp_Arg2), Path)) {
		if (ArchiveEntry->ae_EntryType == ST_FILE || ArchiveEntry->ae_EntryType == ST_LINKFILE) {
			if ((DosPacket->dp_Res2 = GetAccess(O, ArchiveEntry, SHARED_LOCK)) == 0) {
				if (ArchiveHandle = AllocArchiveObjectA(O, AO_Handle, NULL)) {
					ArchiveHandle->ah_FileHandle = FileHandle;
					ArchiveHandle->ah_Entry = ArchiveEntry;
					ArchiveHandle->ah_Mode = MODE_OLDFILE;
					ArchiveHandle->ah_Access = SHARED_LOCK;
					FileHandle->fh_Arg1 = (LONG)ArchiveHandle;
					DosPacket->dp_Res2 = 0;
					DoAAHM_Open(O, ArchiveHandle, &DosPacket->dp_Res2, TAG_END);
					if (DosPacket->dp_Res2 == 0) {
						DosPacket->dp_Res1 = DOSTRUE;
						return 0;
					}
					FreeArchiveObject(O, AO_Entry, ArchiveHandle);
				}
				else
					DosPacket->dp_Res2 = ERROR_NO_FREE_STORE;
				ReleaseAccess(O, ArchiveEntry, SHARED_LOCK);
			}
		}
		else
			DosPacket->dp_Res2 = ERROR_OBJECT_WRONG_TYPE;
	}
	else
		DosPacket->dp_Res2 = ERROR_OBJECT_NOT_FOUND;
	return 0;
}


static ULONG do_Packet_READ(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveHandle *ArchiveHandle;

	/*
	 *	dp_Arg1: FileHandle->fh_Arg1;
	 *	dp_Arg2: void *Buffer
	 *	dp_Arg3: LONG BytesToRead
	 *
	 *	dp_Res1: LONG Bytes read, 0 = EOF, -1 = error
	 *	dp_Res2: LONG error code (if dp_Res1 == -1)
	 */

	DBG_Trace();
	DosPacket->dp_Res1 = -1;
	if (ArchiveHandle = (struct PrivateArchiveHandle *)DosPacket->dp_Arg1) {
		DoAAHM_Read(O, ArchiveHandle, (APTR)DosPacket->dp_Arg2, DosPacket->dp_Arg3,
					   &DosPacket->dp_Res1, &DosPacket->dp_Res2, TAG_END);
	}
	return 0;
}


static ULONG do_Packet_SEEK(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveHandle *ArchiveHandle;
	struct PrivateArchiveEntry *ArchiveEntry;
	LONG NewOffset;

	/*
	 *	dp_Arg1: FileHandle->fh_Arg1;
	 *	dp_Arg2: LONG position or offset
	 *	dp_Arg3: LONG seek mode
	 *
	 *	dp_Res1: LONG absolute position before this seek
	 *	dp_Res2: LONG error code (if dp_Res1 == -1)
	 */

	DBG_Trace();
	DosPacket->dp_Res1 = -1;
	if (ArchiveHandle = (struct PrivateArchiveHandle *)DosPacket->dp_Arg1) {
		ArchiveEntry = (struct PrivateArchiveEntry *)ArchiveHandle->ah_Entry;
		switch (DosPacket->dp_Arg3) {
		case OFFSET_BEGINNING:
			NewOffset = DosPacket->dp_Arg2;
			break;
		case OFFSET_CURRENT:
			NewOffset = ArchiveHandle->ah_Pos + DosPacket->dp_Arg2;
			break;
		case OFFSET_END:
			NewOffset = ArchiveEntry->ae_Size + DosPacket->dp_Arg2;
			break;
		}
		if (NewOffset >= 0 && NewOffset <= ArchiveEntry->ae_Size) {
			DosPacket->dp_Res1 = ArchiveHandle->ah_Pos;
			ArchiveHandle->ah_Pos = NewOffset;
		}
		else
			DosPacket->dp_Res2 = ERROR_SEEK_ERROR;
	}
	else
		DosPacket->dp_Res2 = ERROR_INVALID_LOCK;
	return 0;
}


static ULONG do_Packet_END(Object *O, struct TagItem *TagList, struct DosPacket *DosPacket)
{
	struct PrivateArchiveHandle *ArchiveHandle;

	/*
	 *	dp_Arg1: FileHandle->fh_Arg1;
	 *
	 *	dp_Res1: LONG DOSTRUE/DOSFALSE
	 *	dp_Res2: LONG error code (if dp_Res1 == DOSFALSE)
	 */

	DBG_Trace();
	if (ArchiveHandle = (struct PrivateArchiveHandle *)DosPacket->dp_Arg1) {
		DoAAHM_Close(O, ArchiveHandle, TAG_END);
		ReleaseAccess(O, (struct PrivateArchiveEntry *)ArchiveHandle->ah_Entry, ArchiveHandle->ah_Access);
		FreeArchiveObject(O, AO_Entry, ArchiveHandle);
		DosPacket->dp_Res1 = DOSTRUE;
	}
	return 0;
}


ULONG ahm_DoPacket(Class *Cl, Object *O, struct ahpDoPacket *M)
{
	struct ArchiveData *Data = INST_DATA(Cl, O);
	struct DosPacket *DosPacket = M->Pkt;
	struct TagItem *TagList = M->TagList;
	ULONG RC = NULL;			/* generic return value, meaning depend on the method */

	DBG_Trace();
	DBG_Info(PacketName(DosPacket->dp_Type));
	switch (DosPacket->dp_Type) {
	case ACTION_DISK_INFO:
		RC = do_Packet_DISK_INFO(O, TagList, DosPacket);
		break;
	case ACTION_INFO:
		RC = do_Packet_INFO(O, TagList, DosPacket);
		break;
	case ACTION_FREE_LOCK:
		RC = do_Packet_FREE_LOCK(O, TagList, DosPacket);
		break;
	case ACTION_LOCATE_OBJECT:
		RC = do_Packet_LOCATE_OBJECT(O, TagList, DosPacket);
		break;
	case ACTION_COPY_DIR:
		RC = do_Packet_COPY_DIR(O, TagList, DosPacket);
		break;
	case ACTION_PARENT:
		RC = do_Packet_PARENT(O, TagList, DosPacket);
		break;
	case ACTION_PARENT_FH:
		RC = do_Packet_PARENT_FH(O, TagList, DosPacket);
		break;
	case ACTION_FH_FROM_LOCK:
		RC = do_Packet_FH_FROM_LOCK(O, TagList, DosPacket);
		break;
	case ACTION_EXAMINE_FH:
		RC = do_Packet_EXAMINE_FH(O, TagList, DosPacket);
		break;
	case ACTION_COPY_DIR_FH:
		RC = do_Packet_COPY_DIR_FH(O, TagList, DosPacket);
		break;
	case ACTION_EXAMINE_OBJECT:
		RC = do_Packet_EXAMINE_OBJECT(O, TagList, DosPacket);
		break;
	case ACTION_EXAMINE_NEXT:
		RC = do_Packet_EXAMINE_NEXT(O, TagList, DosPacket);
		break;
	case ACTION_FINDINPUT:
		RC = do_Packet_FINDINPUT(O, TagList, DosPacket);
		break;
	case ACTION_READ:
		RC = do_Packet_READ(O, TagList, DosPacket);
		break;
	case ACTION_SEEK:
		RC = do_Packet_SEEK(O, TagList, DosPacket);
		break;
	case ACTION_END:
		RC = do_Packet_END(O, TagList, DosPacket);
		break;
	case ACTION_IS_FILESYSTEM:
		DosPacket->dp_Res1 = DOSTRUE;
		DosPacket->dp_Res2 = 0;
		break;
	case ACTION_DIE:
		Data->Flags |= ADF_QuitHandler;		// Quit as soon as possible...
		DosPacket->dp_Res1 = DOSTRUE;
		DosPacket->dp_Res2 = 0;
		break;


//TODO
	case ACTION_CURRENT_VOLUME:
	case ACTION_READ_LINK:
	case ACTION_CHANGE_MODE:
		break;

// Write protected...
	case ACTION_WRITE_PROTECT:
	case ACTION_RENAME_DISK:
	case ACTION_DELETE_OBJECT:
	case ACTION_RENAME_OBJECT:
	case ACTION_SET_PROTECT:
	case ACTION_CREATE_DIR:
	case ACTION_SET_COMMENT:
	case ACTION_SET_DATE:
	case ACTION_MAKE_LINK:
	case ACTION_SET_OWNER:
	case ACTION_FINDUPDATE:
	case ACTION_FINDOUTPUT:
	case ACTION_SET_FILE_SIZE:
		DosPacket->dp_Res1 = DOSFALSE;
		DosPacket->dp_Res2 = ERROR_DISK_WRITE_PROTECTED;
		break;
	case ACTION_WRITE:
		DosPacket->dp_Res1 = -1;
		DosPacket->dp_Res2 = ERROR_DISK_WRITE_PROTECTED;
		break;
	}
	return RC;
}

