
#ifndef ARCHIVECLASS_H
#define ARCHIVECLASS_H

#define ARCHIVECLASS	"archiveclass"

/*
 *	Archive Methods.
 *	Prefix rules:
 *	AM_...		Archive Method (application context execution)
 *	AHM_...		Archive Handler Method (handler context execution)
 *	AAHM_...	Archive Asynchronous Handler Method (Archive's slave process context execution)
 *	AAVHM_...	Archive Asynchronous Virtual Handler Method (Archive's slave process context execution)
 */

#define AM_Dummy 				(TAG_USER)
#define AM_Mount				(AM_Dummy + 0x00)		/* Object *. Mount archive. Returns RootLock. */
#define AM_Add					(AM_Dummy + 0x01)		/* ? */
#define AM_Extract				(AM_Dummy + 0x02)		/* ? */
#define AM_List					(AM_Dummy + 0x03)		/* ? */

#define AHM_Dummy 				(TAG_USER + 0x40)
#define AHM_Startup				(AHM_Dummy + 0x00)		/* Init handler (in handler context) */
#define AHM_Finish				(AHM_Dummy + 0x01)		/* Finish handler (in handler context) */
#define AHM_DoPacket			(AHM_Dummy + 0x02)		/* struct opAHM_DoPacket * */


#define AAHM_Dummy 				(TAG_USER + 0x80)
#define AAHM_ScanArchive		(AAHM_Dummy + 0x00)		/* */
#define AAHM_WaitEndScan		(AAHM_Dummy + 0x01)		/* */

/*
 *  Virtual methods which must be supported by subclasses if AM_Mount is supported
 */
#define AAHM_ReadFirstHeader	(AAHM_Dummy + 0x40)
#define AAHM_ReadNextHeader		(AAHM_Dummy + 0x41)
#define AAHM_ReadHeaderEnd		(AAHM_Dummy + 0x42)
#define AAHM_Open				(AAHM_Dummy + 0x43)
#define AAHM_Read				(AAHM_Dummy + 0x44)
#define AAHM_Write				(AAHM_Dummy + 0x45)
#define AAHM_Close				(AAHM_Dummy + 0x46)


/* Tags for MountArchive() (& AM_Mount method)  */
#define MA_Dummy				(TAG_USER + 0x800)
#define MA_AutoUnmount			(MA_Dummy + 0)			/* (I) BOOL         */
#define MA_DeviceName			(MA_Dummy + 1)			/* (G) UBYTE *      */


/* Tags for NewArchive() (& OM_NEW method)  */
#define NA_Dummy				(TAG_USER + 0x900)
#define NA_FileType				(NA_Dummy + 0)			/* (I) ULONG        */
#define NA_Create				(NA_Dummy + 1)			/* (I) BOOL Create if doesn't exists  */


struct ahpDoPacket {
	ULONG MethodID;
	struct DosPacket *Pkt;
	struct TagItem *TagList;
};

struct aahpReadHeader {
	ULONG MethodID;
	char DirPath[256];
	struct ArchiveEntry *Entry;		/* Allocate and fill in */
	struct TagItem *TagList;
	ULONG Error;
	APTR UserData;
};

struct aahpOpen {
	ULONG MethodID;
	struct ArchiveHandle *ArchiveHandle;
	LONG *Error;
	struct TagItem *TagList;
};

struct aahpRead {
	ULONG MethodID;
	struct ArchiveHandle *ArchiveHandle;
	APTR Buffer;
	LONG BytesToRead;
	LONG *BytesRead;
	LONG *Error;
	struct TagItem *TagList;
};

struct aahpClose {
	ULONG MethodID;
	struct ArchiveHandle *ArchiveHandle;
	struct TagItem *TagList;
};

struct opTagList {
	ULONG MethodID;
	struct TagItem *TagList;
};


/* Archive Attributes */

#define AA_Dummy 				(TAG_USER + 0x1000)
#define AA_ArchiveFileName		(AA_Dummy + 0x01)		/* (I) UBYTE * 				*/
#define AA_DeviceBaseName		(AA_Dummy + 0x02)		/* (G) UBYTE * 				*/
#define AA_VolumeName			(AA_Dummy + 0x03)		/* (G) UBYTE * 				*/
#define AA_DeviceName			(AA_Dummy + 0x04)		/* (G) UBYTE * 				*/
#define AA_VolumeInfo			(AA_Dummy + 0x05)		/* (G) struct InfoData *	*/


/* Archive Objects (alloc with AllocHandlerObject()) */

/*
 *	ASync Message. you MUST allocate with AllocArchiveObject()
 *	It can be extented just like any other message. Just do:
 *	MyMsg = AllocArchiveObject(O, AO_ASyncRequest, AAO_ObjectSize, sizeof(struct MyMessage), TAG_END);
 */
struct ASyncRequest {
	struct Message asm_Message;
};


struct ArchiveEntry {
	struct MinNode ae_Node;			/* PRIVATE to archive class! */

	/* Public part. Init this field when ReadHeader() */
	UBYTE  ae_Name[32];						/* Filename         */
	WORD   ae_EntryType;					/* fib_DirEntryType */
	UWORD  ae_Protection;					/* Protection bits  */
	LONG   ae_Size;							/* Size in bytes    */
	LONG   ae_PackedSize;					/* In bytes too     */
	LONG   ae_NumBlocks;					/* Size in blocks   */
	struct DateStamp ae_Date;				/* File Date        */
	UBYTE  *ae_Comment;						/* File Comment     */
	UBYTE  *ae_Link;						/* Link Path		*/
	UWORD  ae_OwnerUID;						/* owner's UID      */
	UWORD  ae_OwnerGID;						/* owner's GID      */
	struct MinList ae_SubEntries;			/* For dirs only    */
	struct PrivateArchiveEntry *ae_Parent;	/* Parent dir       */
	/* Private data follows. Hands OFF! ;-) */
};


struct ArchiveHandle {
	struct FileHandle *ah_FileHandle;	/* Standard DOS FileHandle */
	struct ArchiveEntry *ah_Entry;		/* The entry of this file */
	LONG ah_Mode;						/* Open() mode */
	LONG ah_Pos;						/* Seek pos. FileHandle->fh_Pos is reserved for dos */
	/* Private data follows. Hands OFF! ;-) */
};


/* Archive Objects (alloc with AllocHandlerObject()) */
#define AO_Custom		0		// MUST give size (AOA_ObjectSize) too
#define AO_ASyncRequest	1
#define AO_Entry		2
#define AO_Lock			3
#define AO_Handle		4


/* Tags for AllocArchiveObject() */
#define AAO_ObjectSize		(TAG_USER)			// Only valid for AO_Custom & AO_ASyncRequest



/* Extension of the Class structure to store archive subobjects information for INST_... macros */
/* DO NOT ALLOCATE BY YOURSELF. This struct may grow in size. Use InitArchiveClassData().
/* Currently only for 'struct Entry' and 'struct Handle' */

struct ArchiveClassData {
	APTR UserData;				/* Use this in place of the cl_UserData which is used to store this struct */
	ULONG EntrySize;			/* Don't worry about these fields. Uses following macros. */
	ULONG EntryOffset;
	ULONG HandleSize;
	ULONG HandleOffset;
};

/* Tags for InitArchiveClassData() */
#define ACD_Dummy			(TAG_USER)
#define ACD_EntrySize		(ACD_Dummy + 1)
#define ACD_HandleSize		(ACD_Dummy + 2)


#define INST_EntryData(Cl, Entry)	((void *)((UBYTE *)Entry + ((struct ArchiveClassData *)Cl->cl_UserData)->EntryOffset))
#define INST_HandleData(Cl, Handle)	((void *)((UBYTE *)Handle + ((struct ArchiveClassData *)Cl->cl_UserData)->HandleOffset))


/* Archive definition file (located in DEVS:Archives) */

#define ID_ARCV		MAKE_ID('A', 'R', 'C', 'V')
#define ID_ARCT		MAKE_ID('A', 'R', 'C', 'T')

struct ArchiveType {
	UBYTE ObjectBaseName[24];
	UBYTE FileTypeName[24];
	ULONG Flags;
};

#endif

