/*
 *  Lha.c - Copyright ) 1994 by J.B.Y.
 *
 *  Created:    26 Dec 1994
 *  Modified:   
 *
 *  Make>> smake
 */

#include <clib.h>
#include "Debug.h"

#include <proto/Archive.h>
#include <archives/ArchiveClass.h>
#include <archives/LhaClass.h>
#include <archives/ArchiveMethods.h>
#include <archives/LhaMethods.h>

#include "Lha.h"
#include "proto/Lha.h"


static ULONG OM_New(Class *Cl, Object *O, struct opSet const *oS)
{
	struct opSet opSet = *oS;
	ULONG RC = NULL;			/* generic return value, meaning depend on the method */
	UBYTE *FileName;
	BPTR L;

	DBG_Trace();
	if (!(FileName = (UBYTE *)GetTagData(AA_ArchiveFileName, NULL, opSet.ops_AttrList)))
		return NULL;
	if (!(L = Lock(FileName, ACCESS_READ)))
		return NULL;
	if (RC = DoSuperMethodA(Cl, O, (Msg)&opSet)) {
		__aligned struct FileInfoBlock fib;
		const Object *ThisObject = (Object *)RC;
		struct LhaData *Data = INST_DATA(Cl, ThisObject);

		memset(Data, 0, sizeof(struct LhaData));
		Data->LhaFileLock = L;
		Examine(L, &fib);
		strcpy(Data->LhaFileName, fib.fib_FileName);
		Data->LhaFileSize = fib.fib_Size;
	}
	else
		UnLock(L);
	return RC;
}


static ULONG OM_Dispose(Class *Cl, Object *O, Msg M)
{
	struct LhaData *const Data = INST_DATA(Cl, O);

	if (Data->LhaFileLock);
		UnLock(Data->LhaFileLock);
	return DoSuperMethodA(Cl, O, M);
}


static ULONG OM_Set(Class *Cl, Object *O, struct opSet * const oS)
{
/*
	ULONG RC = 0;
	struct TagItem *i, *j;

	DBG_Trace();
	for (j = oS->ops_AttrList, i = NextTagItem(&j); i; i = NextTagItem(&j)) {
		switch (i->ti_Tag) {
		case 0:
			break;
		}
	}
*/
	return DoSuperMethodA(Cl, O, (Msg) oS);
}


static void FillInfoData(struct LhaData *Data, struct InfoData *InfoData)
{
	#define BLOCK_SIZE	512

	InfoData->id_DiskState = ID_WRITE_PROTECTED;	/* Default setting. You may change this if you support write */
	InfoData->id_NumBlocks = (Data->LhaFileSize + (BLOCK_SIZE - 1)) / BLOCK_SIZE;
	if (InfoData->id_NumBlocks == 0)
		InfoData->id_NumBlocks = 1;		// Prevent zero divide in apps that does not check it.
	InfoData->id_NumBlocksUsed = InfoData->id_NumBlocks;
	InfoData->id_BytesPerBlock = BLOCK_SIZE;
}


static ULONG OM_Get(Class *Cl, Object *O, struct opGet *const oG)
{
	struct LhaData *const Data = INST_DATA(Cl, O);
	ULONG RC = 0;

	switch (oG->opg_AttrID) {
	case AA_DeviceBaseName:
		*oG->opg_Storage = (ULONG)"LHA";
		break;
	case AA_VolumeName:
		*oG->opg_Storage = (ULONG)Data->VolumeName;
		break;
	case AA_VolumeInfo:
		FillInfoData(Data, (struct InfoData *)oG->opg_Storage);
		break;
	default:
		RC = DoSuperMethodA(Cl, O, (Msg) oG);
	}
	return RC;
}


void ahm_Finish(Class *Cl, Object *O, struct opTagList *M)
{
	struct LhaData * const Data = INST_DATA(Cl, O);

	DBG_Trace();
	if (Data->LhaFH)
		DoTHM_CloseLhaFile(O, Data->LhaFH, TAG_END);
	DoSuperMethodA(Cl, O, (Msg)M);
}


BPTR am_Mount(Class *Cl, Object *O, struct opTagList *M)
{
	struct LhaData * const Data = INST_DATA(Cl, O);
	short len;
	char *p;

	strcpy(Data->VolumeName, Data->LhaFileName);
	#ifdef DEBUG
	Data->HandlerProc = SysBase->ThisTask;			// Debug purpose
	#endif
	if ((len = strlen(Data->VolumeName)) > 4) {
		p = Data->VolumeName + len - 4;
		if (!Stricmp(p, ".lha"))
			*p = '\0';
	}
	return (BPTR)DoSuperMethodA(Cl, O, (Msg)M);
}


static ULONG __saveds __asm far Dispatch(register __a0 Class *Cl, register __a2 Object *O, register __a1 Msg M)
{
	ULONG RC = NULL;			/* generic return value, meaning depend on the method */

	if ((M->MethodID & AAHM_Dummy) == AAHM_Dummy) {
		#ifdef DEBUG
		if (SysBase->ThisTask == Data->HandlerProc) {
			SimpleReq("ERROR! ERROR! ERROR! ERROR! ERROR!\n!!!! ASynchronous method called in handler context !!!!");
			return 0;
		}
		#endif	
		switch (M->MethodID) {
		case AAHM_ReadFirstHeader:
			RC = aahm_ReadFirstHeader(Cl, O, (struct aahpReadHeader *)M);
			break;
		case AAHM_ReadNextHeader:
			RC = aahm_ReadNextHeader(Cl, O, (struct aahpReadHeader *)M);
			break;
		case AAHM_ReadHeaderEnd:
			RC = aahm_ReadHeaderEnd(Cl, O, (struct aahpReadHeader *)M);
			break;
		case AAHM_Open:
			RC = aahm_Open(Cl, O, (struct aahpOpen *)M);
			break;
		case AAHM_Read:
			RC = aahm_Read(Cl, O, (struct aahpRead *)M);
			break;
		case AAHM_Write:
			break;
		case AAHM_Close:
			RC = aahm_Close(Cl, O, (struct aahpClose *)M);
			break;
		case TAHM_OpenLhaFile:
			RC = (ULONG)tahm_OpenLhaFile(Cl, O, (struct tahpOpenLhaFile *)M);
			break;
		default:
			RC = DoSuperMethodA(Cl, O, M);
			break;
		}
	}
	else {
		switch (M->MethodID) {
		case AHM_DoPacket:
			RC = DoSuperMethodA(Cl, O, M);
			break;
		case OM_NEW:
			RC = OM_New(Cl, O, (struct opSet * const) M);
			break;
		case OM_DISPOSE:
			RC = OM_Dispose(Cl, O, M);
			break;
		case OM_GET:
			RC = OM_Get(Cl, O, (struct opGet * const) M);
			break;
		case OM_SET:
			RC = OM_Set(Cl, O, (struct opSet * const) M);
			break;
		case AM_Mount:
			RC = am_Mount(Cl, O, (struct opTagList *)M);
			break;
		case AHM_Finish:
			ahm_Finish(Cl, O, (struct opTagList *)M);
			break;
		case THM_CloseLhaFile:
			thm_CloseLhaFile(Cl, O, (struct thpCloseLhaFile *)M);
			break;
		default:
			RC = DoSuperMethodA(Cl, O, M);
		}
	}
	DBG_Info1(KPrintF("%s, %s(), %l04d>Return:$%l04x\n", __FILE__, __FUNC__, __LINE__, RC));
	return (ULONG) RC;
}


Class *InitCl(void)
{
	Class *Cl;

	DBG_Trace();
	if (Cl = MakeClass(LHACLASS, ARCHIVECLASS, NULL, sizeof(struct LhaData), 0)) {
		if (InitArchiveClassData(Cl, ACD_EntrySize, sizeof(struct LhaEntry),
									 ACD_HandleSize, sizeof(struct LhaHandle),
									 TAG_END))
		{
			Cl->cl_Dispatcher.h_Entry = (ULONG(*)())Dispatch;
			Cl->cl_Dispatcher.h_SubEntry = NULL;
			//Cl->cl_Dispatcher.h_Data = &TarData;       /* not neccsary in a Class */
			AddClass(Cl);
			return Cl;
		}
		FreeClass(Cl);
	}
	return NULL;
}


BOOL FreeCl(Class *Cl)
{
	DBG_Trace();
	if (Cl) {
		RemoveClass(Cl);			/* NULL ptr on private class is legal */
		FreeArchiveClassData(Cl);
		return FreeClass(Cl);
	}
	return TRUE;
}

