#include "LHAUtils.h"

unsigned char	*convdelim(unsigned char *path,unsigned char delim)
{
unsigned char	*p;

	for (p=path; *p; p++)
		if (*p=='\\' || *p==DELIM || *p==DELIM2)
			*p = delim;
	return path;
}

/*----------------------------------------------------------------------*/
/*                                                                      */
/*      Generic stamp format:                                           */
/*                                                                      */
/*       31 30 29 28 27 26 25 24 23 22 21 20 19 18 17 16                */
/*      |<-------- year ---->|<- month ->|<--- day ---->|               */
/*                                                                      */
/*       15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0                */
/*      |<--- hour --->|<---- minute --->|<- second/2 ->|               */
/*                                                                      */
/*----------------------------------------------------------------------*/
#define SECOND_SHIFT	0
#define SECOND_MASK	0x1f	/* No more than 2*32 secondes per minute */
#define MINUTE_SHIFT	5
#define MINUTE_MASK	0x3f	/* No more than 64 minutes per hour */
#define HOUR_SHIFT	11
#define HOUR_MASK	0x1f	/* No more than 32 hours per day */
#define DAY_SHIFT	16
#define DAY_MASK	0x1f	/* No more than 32 days per month */
#define MONTH_SHIFT	21
#define MONTH_MASK	0x0f	/* No more than 16 month per year */
#define YEAR_SHIFT	25
#define YEAR_MASK	0x7f	/* No more than 128 years per AMIGA-life */

void	convert_time(int type,long time,time_t *unix_time)
{
static int	from0jan[12] = { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273,
			304, 334 };
long		year, month, day, hour, min, sec, days;
	switch (type) {
		case GENERIC_TO_UNIX:
			year = ((time>>YEAR_SHIFT)&YEAR_MASK)+1980;
			month = (time>>MONTH_SHIFT)&MONTH_MASK;
			day = (time>>DAY_SHIFT)&DAY_MASK;
			hour = (time>>HOUR_SHIFT)&HOUR_MASK;
			min = (time>>MINUTE_SHIFT)&MINUTE_MASK;
			sec = ((time>>SECOND_SHIFT)&SECOND_MASK)*2;
			days = (365*(year-1970))+
				(year-1970+1)/4+
				from0jan[month-1]+
				day-1;
			if ((year%4==0) && (year%400!=0) && (month>=3))
				days++;
			*unix_time = (((days*24)+hour)*60+min)*60+sec;
			if (time==0) *unix_time = 0;
			break;
	}
	return;
}

void	convert_name(int type,char name[256],int length)
{
int	i, lower_used;

        length = strlen(name);
       	switch(type) {
		case GENERIC_TO_UNIX:
                        lower_used = FALSE;
			for (i=0; i<length; i++) {
				if (name[i]=='\\') name[i] = '/';
				if (islower(name[i])) lower_used = TRUE;
			}
			if (!lower_used)
				for (i=0; i<length; i++)
					if (isupper(name[i]))
						name[i] = tolower(name[i]);
			break;
		case MACOS_TO_UNIX:
			for (i=0; i<length; i++)
				if (name[i]==':') name[i] = '/';
				else if (name[i]=='/') name[i] = ':';
			break;
		case MSDOS_TO_UNIX:
			for (i=0; i<length; i++)
				if (name[i]=='\\') name[i] = '/';
				else if (isupper(name[i]))
					name[i] = tolower(name[i]);
			break;
	}
}

int	calc_sum(char *p,int l)
{
int	sum;

	for(sum=0; l; l--) sum += *p++;
	return sum & 0xff;
}

int		decode8(unsigned char *p)
{
	return *p & 0xff;
}

int		decode16(unsigned char *p)
{
	return (decode8(p+1)<<8)+decode8(p);
}

int		decode32(unsigned char *p)
{
	return (decode8(p+3)<<24)+(decode8(p+2)<<16)+
		(decode8(p+1)<<8)+decode8(p);
}

char	*get_stamp_string(time_t stamp)
{
static char	month[12*3+1] = "JanFebMarAprMayJunJulAugSepOctNovDec",
		buffer[256];
struct tm	*p;
time_t		now;
long		threshold;

	if (stamp==(time_t)0) return "                 ";
/*
	now = time((time_t *)NULL);
	p = localtime(&now);
	threshold = p->tm_year*12+p->tm_mon-6;
	p = localtime(&stamp);
	if (p->tm_year*12+p->tm_mon>threshold)
		sprintf(buffer,"%.3s %2d %02d:%02d",&month[p->tm_mon*3],
			p->tm_mday,p->tm_hour,p->tm_min);
	else
		sprintf(buffer,"%.3s %2d  %04d",&month[p->tm_mon*3],
			p->tm_mday,p->tm_year+1900);
	return buffer;
*/
	return "                 ";
}

ERROR		read_header(BPTR file,LzHeader *header,int *read_size)
{
char		buffer[LZHEADER_STRAGE], dirname[256], *pbuf, *old_pbuf;
int		header_size=0, l=0, checksum=0, calc_checksum=0, name_length=0,
		i=0, dir_length=0, ii;

	/* Lit la taille du header */
	*read_size = l = Read(file,buffer+I_HEADER_SIZE,(long)1);
	if (l!=1)				return SHORT_HEADER;
	header_size = buffer[I_HEADER_SIZE];
	if (header_size==EOF || header_size==0)	return END_OF_ARCHIVE;
	if (header_size<0)			return BAD_HEADER_SIZE;
	header->header_size = header_size;

	/* Lit le header */
	l = Read(file,buffer+I_HEADER_CHECKSUM,(long)(header_size-1));
	if (l!=header_size-1)			return SHORT_HEADER;
	*read_size += l;

	/* Quel niveau ? */
	header->header_level = decode8(buffer+I_HEADER_LEVEL);
	if (header->header_level!=HEADER_LEVEL2) {
		l = Read(file,buffer+header_size,(long)2);
		*read_size += l;
		if (l!=2)			return SHORT_HEADER;
	}

	/* Prends la checksum */
	checksum = decode8(buffer+I_HEADER_CHECKSUM);

	/* Extrait la methode */
	for (ii=0; ii<METHOD_TYPE_STRAGE; ii++)
		((char *)(header->method))[ii] = buffer[I_METHOD+ii];
/*
	bcopy(buffer+I_METHOD,header->method,METHOD_TYPE_STRAGE);
*/

	/* Decode les champs du header */
	header->packed_size = decode32(buffer+I_PACKED_SIZE);
	header->original_size = decode32(buffer+I_ORIGINAL_SIZE);
	header->last_modified_stamp = decode32(buffer+I_LAST_MODIFIED_STAMP);
	header->attribute = decode8(buffer+I_ATTRIBUTE);

	if (header->header_level!=HEADER_LEVEL2) {
		/* Verifie la checksum */
		calc_checksum = calc_sum(buffer+I_METHOD,header_size);
		if (checksum != calc_checksum)	return BAD_HEADER_CHECKSUM;
		name_length = decode8(buffer+I_NAME_LENGTH);
		for (i=0; i<name_length; i++)
			header->name[i] = (buffer+I_NAME)[i];
		header->name[name_length] = '\0';
	}
	else {
		header->unix_last_modified_stamp = header->last_modified_stamp;
		name_length = 0;
	}

	/* valeurs par defaut pour les champs speciaux */
	header->unix_mode = UNIX_FILE_REGULAR | UNIX_RW_RW_RW;
	header->unix_gid = 0;
	header->unix_uid = 0;

	/* Quel format etendu ? */
	if (header_size - name_length >= 24) {	/* EXTEND FORMAT */
		header->crc = decode16(buffer+I_CRC+name_length);
		header->extend_type = decode8(buffer+I_EXTEND_TYPE+name_length);
		header->minor_version = decode8(buffer+I_MINOR_VERSION+
			name_length);
		header->has_crc = TRUE;
	}
	else if (header_size - name_length == 22) {	/* GENERIC WITH CRC */
		header->crc = decode16(buffer+I_CRC+name_length);
		header->extend_type = EXTEND_GENERIC;
		header->has_crc = TRUE;
        }
	else if (header_size - name_length == 20) {	/* GENERIC NO CRC */
		header->extend_type = EXTEND_GENERIC;
		header->has_crc = FALSE;
	}
	else					return UNKNOWN_HEADER_TYPE;

	if (header->extend_type==EXTEND_UNIX &&
	    header->header_level==HEADER_LEVEL0) {
		header->minor_version = decode8(buffer+I_MINOR_VERSION);
		header->unix_last_modified_stamp =
			(time_t)decode32(buffer+I_UNIX_LAST_MODIFIED_STAMP);
		header->unix_mode = decode16(buffer+I_UNIX_MODE);
		header->unix_uid = decode16(buffer+I_UNIX_UID);
		header->unix_gid = decode16(buffer+I_UNIX_GID);
						return NO_ERROR;
	}

	if (header->header_level > HEADER_LEVEL0) {	/* EXTEND HEADER */
		if (header->header_level!=HEADER_LEVEL2)
			pbuf = buffer+header->header_size;
		old_pbuf = pbuf;
		while ((header_size=decode16(pbuf))!=0) {
			pbuf += 2;
			if (header->header_level!=HEADER_LEVEL2) {
				if (buffer+LZHEADER_STRAGE-pbuf < header_size)
						return BAD_HEADER_SIZE;
				l = Read(file,pbuf,(long)header_size);
				*read_size += l;
				if (l<header_size)
						return BAD_HEADER_SIZE;
			}
			switch (decode8(pbuf)) {
				case 0:	/* HEADER_CRC */
					pbuf++;
					pbuf += header_size-3;
					break;
				case 1:	/* FILENAME */
					pbuf++;
					for (i=0; i<header_size-3; i++)
						header->name[i] = decode8(pbuf+i);
					header->name[header_size-3] = '\0';
					pbuf += header_size-3;
					break;
				case 2:	/* DIRECTORY */
					pbuf++;
					for (i=0; i<header_size-3; i++)
						dirname[i] = decode8(pbuf+i);
					dirname[header_size-3]='\0';
					convdelim(dirname,DELIM);
					dir_length = header_size-3;
					pbuf += header_size-3;
					break;
				case 0x40:	/* MS/DOS ATTRIBUTE */
					pbuf++;
					if (header->extend_type==EXTEND_MSDOS ||
					    header->extend_type==EXTEND_HUMAN ||
					    header->extend_type==EXTEND_GENERIC)
						header->attribute =decode8(pbuf);
					pbuf++;
					break;
				case 0x50:	/* UNIX PERMISSION */
					pbuf++;
					if (header->extend_type==EXTEND_UNIX) {
						header->unix_mode = decode16(pbuf);
						pbuf += 2;
					}
					break;
				case 0x51:	/* UNIX GID & UID */
					pbuf++;
					if (header->extend_type==EXTEND_UNIX) {
						header->unix_gid = decode16(pbuf);
						pbuf += 2;
						header->unix_uid = decode16(pbuf);
						pbuf += 2;
					}
					break;
				case 0x52:	/* UNIX GROUP NAME */
					pbuf++;
					pbuf += header_size-3;
					break;
				case 0x53:	/* UNIX USER NAME */
					pbuf++;
					pbuf += header_size-3;
					break;
				case 0x54:	/* UNIX LAST MODIFIED TIME */
					pbuf++;
					if (header->extend_type==EXTEND_UNIX) {
						header->unix_last_modified_stamp
							= decode32(pbuf);
						pbuf += 4;
					}
					break;
				default:	/* OTHER HEADERS */
					pbuf++;
					pbuf += header_size-3;
					break;
			}
		}
		pbuf += 2;
		if (header->header_level!=HEADER_LEVEL2 && pbuf-old_pbuf!=2) {
			header->packed_size -= pbuf-old_pbuf-2;
			header->header_size += pbuf-old_pbuf-2;
		}
	}
	if (dir_length) {
		strcat(dirname,header->name);
		strcpy(header->name,dirname);
		name_length += dir_length;
	}
	switch (header->extend_type) {
		case EXTEND_MSDOS:
			/* MS/DOS to UNIX filename conversion */
			convert_name(MSDOS_TO_UNIX,header->name,name_length);
		case EXTEND_HUMAN:
			if (header->header_level==2)
				header->unix_last_modified_stamp =
					header->last_modified_stamp;
			else
			/* Generic to UNIX time stamp conversion */
				convert_time(GENERIC_TO_UNIX,
					header->last_modified_stamp,
					&header->unix_last_modified_stamp);
			break;
		case EXTEND_OS68K:
		case EXTEND_XOSK:
		case EXTEND_UNIX:
			break;
		case EXTEND_MACOS:
			/* MAC/OS to UNIX filename conversion */
			convert_name(MACOS_TO_UNIX,header->name,name_length);
			/* Generic to UNIX time stamp conversion */
			convert_time(GENERIC_TO_UNIX,
				header->last_modified_stamp,
				(time_t *)&header->unix_last_modified_stamp);
			break;
		default:
			/* Generic to UNIX filename conversion */
			convert_name(GENERIC_TO_UNIX,header->name,name_length);
			if (header->header_level==2)
				header->unix_last_modified_stamp =
					header->last_modified_stamp;
			else
				/* Generic to UNIX time stamp conversion */
				convert_time(GENERIC_TO_UNIX,
					header->last_modified_stamp,
       					&header->unix_last_modified_stamp);
	        	break;
	}
	return NO_ERROR;
}
