/*
 *  Header.c - Copyright  1994 by S.R. & P.C.
 *
 *  Created:    30 Nov 1994  23:42:40
 *  Modified:   03 Dec 1994  22:47:43
 *
 *  Make>> smake
 */

#include <clib.h>
#include "Debug.h"

#include <proto/Archive.h>
#include <archives/ArchiveClass.h>
#include <archives/TarClass.h>
#include <archives/ArchiveMethods.h>
#include <archives/TarMethods.h>

#include "Tar.h"
#include "proto/Header.h"


#define HEADERSIZE 512		/* not the size of the struct, but this still the size of the header ! */
#define NAMESIZE 100
#define TUNAMELEN 32
#define TGNAMELEN 32

struct TarHeader {
	char FileName[NAMESIZE];
	char Mode[8];
	char UID[8];
	char GID[8];
	char Size[12];		/* in octal  */
	char mtime[12];
	char CheckSum[8];
	char LinkFlag;
	char Archive_LinkName[NAMESIZE];
	char Magic[8];
	char UserName[TUNAMELEN];
	char GroupName[TGNAMELEN];
	char DevMajor[8];
	char DevMinor[8];
};


struct ReadHeaderContext {
	UBYTE Buffer[HEADERSIZE];
};


static ULONG OctToL(UBYTE *s, UBYTE Len)
{
	ULONG n = 0;

	// skip leading spaces
	while (*s == ' ' && Len--)
		s++;
	while (*s && *s >= '0' && *s <= '7' && Len--) {
		n = n << 3 | *s - '0';
		s++;
	}
	return n;
}


static ULONG ReadHeader(Class *Cl, Object *O, struct aahpReadHeader *M)
{
	struct TarData *const Data = INST_DATA(Cl, O);
	struct ArchiveEntry *ArchiveEntry;
	struct TarEntry *TarEntry;
	struct TarHeader *Header;
	UBYTE l;

	Header = (struct TarHeader *)((struct ReadHeaderContext *)M->UserData)->Buffer;
	if (Read(Data->TarFH, Header, HEADERSIZE) == HEADERSIZE && Header->FileName[0]) {
		if (ArchiveEntry = AllocArchiveObjectA(O, AO_Entry, NULL)) {
			M->Entry = ArchiveEntry;
			l = strlen(Header->FileName);
			if (Header->FileName[l-1] == '/' || Header->FileName[l-1] == ':') {
				Header->FileName[l-1] = '\0';
				ArchiveEntry->ae_EntryType = ST_USERDIR;
			}
			else
				ArchiveEntry->ae_EntryType = ST_FILE;
			strncpy(ArchiveEntry->ae_Name, FilePart(Header->FileName), 31);
			strcpy(M->DirPath, Header->FileName);
			*PathPart(M->DirPath) = '\0';		// remove entry name. just keep parent dir path.
			ArchiveEntry->ae_Size = OctToL(Header->Size, 12);
			ArchiveEntry->ae_PackedSize = ArchiveEntry->ae_Size;
			ArchiveEntry->ae_NumBlocks = (ArchiveEntry->ae_Size + 511) / 512;	/* round up */

			TarEntry = INST_EntryData(Cl, ArchiveEntry);
			TarEntry->te_Offset = Seek(Data->TarFH, ArchiveEntry->ae_NumBlocks * 512, OFFSET_CURRENT);
		}
		else
			M->Error = ERROR_NO_FREE_STORE;
	}
	else
		M->Error = ERROR_NO_MORE_ENTRIES;
	return TRUE;
}


ULONG aahm_ReadFirstHeader(Class *Cl, Object *O, struct aahpReadHeader *M)
{
	struct TarData *const Data = INST_DATA(Cl, O);

	if (M->UserData = AllocVec(sizeof(struct ReadHeaderContext), MEMF_PUBLIC|MEMF_CLEAR)) {
		if (Data->TarFH = DoTAHM_OpenTarFile(O, Data->TarFileLock, TAG_END))
			return ReadHeader(Cl, O, M);
		else
			M->Error = IoErr();
	}
	else
		M->Error = ERROR_NO_FREE_STORE;
	return TRUE;
}


ULONG aahm_ReadNextHeader(Class *Cl, Object *O, struct aahpReadHeader *M)
{
	return ReadHeader(Cl, O, M);
}


ULONG aahm_ReadHeaderEnd(Class *Cl, Object *O, struct aahpReadHeader *M)
{
	if (M->UserData)
		FreeVec(M->UserData);
	return TRUE;
}


BPTR tahm_OpenTarFile(Class *Cl, Object *O, struct tahpOpenTarFile *M)
{
	BPTR fl, fh = NULL;

	if (fl = DupLock(M->FileLock)) {		// Duplicate lock which will be 'absorbed' by OpenFromLock()
		if (!(fh = OpenFromLock(fl)))
			UnLock(fl);
	}
	return fh;
}


void thm_CloseTarFile(Class *Cl, Object *O, struct thpCloseTarFile *M)
{
	if (M->FH)
		Close(M->FH);
}

