#include <exec/types.h>
#include <exec/devices.h>
#include <exec/errors.h>
#include <exec/execbase.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/filehandler.h>
#include <dos/dostags.h>
#include <utility/tagitem.h>
#include <proto/exec.h>
#include <proto/dos.h>
#include <string.h>

#define __USE_SYSBASE

void		ReplyPacket(struct DosPacket *,LONG,LONG);
char		*ToBStr(char *);
struct FileLock	*MyLock(BSTR,struct FileLock *,LONG,struct MsgPort *,
			struct DeviceList *);

struct MyStartup {
	struct MsgPort	*ms_Port;
};

extern struct ExecBase *__far	AbsExecBase;
struct ExecBase		*SysBase;

void __saveds start(void)
{
struct DosLibrary	*DosBase;
struct MsgPort		*mp = NULL;
struct DosPacket	*dp = NULL;
struct DeviceNode       *dn = NULL;
struct DeviceList	*dl = NULL;
struct DosInfo		*di = NULL;
struct Message		*mn = NULL;
struct InfoData		*id = NULL;
struct FileInfoBlock	*fib = NULL;
struct FileLock		*fl = NULL;
struct DeviceList	*myvolume;
struct MsgPort		*dosioport;
int			cont;

        SysBase = AbsExecBase;
	DosBase = (struct DosLibrary *)OpenLibrary(DOSNAME,LIBRARY_MINIMUM);

	/* Get my Message Port */
	mp = &((struct Process *)FindTask(NULL))->pr_MsgPort;

	/* Wait for the Start-up Packet */
	WaitPort(mp);

	/* Now get the Start-up Packet */
	dp = (struct DosPacket *)GetMsg(mp)->mn_Node.ln_Name;

	/* dp->dp_Arg1 : BPTR to BSTR (device name) */
	/* dp->dp_Arg2 : dn->dn_Startup */
	/* dp->dp_Arg3 : BPTR to struct DeviceNode */
	/* dp->dp_Arg4 : struct MsgPort * (alternate handler port) */
        
	/* Get the DeviceNode */
        dn = BADDR(dp->dp_Arg3);
	if (dn!=NULL) {
	        dn->dn_Task = mp;
	}

	/* Get the Dos I/O processor Message Port */
	dosioport = ((struct MyStartup *)BADDR(dp->dp_Arg2))->ms_Port;

	/* Now create the Volume node */
	myvolume = (struct DeviceList *)AllocVec(sizeof(struct DeviceList),
			MEMF_CLEAR|MEMF_CHIP);
	if (myvolume==NULL) {
		dn->dn_Task = NULL;
		ReplyPacket(dp,DOSFALSE,ERROR_NO_FREE_STORE);
		return;
	}
	myvolume->dl_Type = DLT_VOLUME;
	myvolume->dl_Task = mp;
	myvolume->dl_Lock = 0;
	//myvolume->dl_VolumeDate =
	myvolume->dl_LockList = NULL;
	myvolume->dl_DiskType = ID_DOS_DISK;;
	//myvolume->dl_unused =
	myvolume->dl_Name = (BSTR)MKBADDR(ToBStr("LHA"));

	/* Now link it!!! */
	di = (struct DosInfo *)BADDR(DosBase->dl_Root->rn_Info);
	myvolume->dl_Next = di->di_DevInfo;
	di->di_DevInfo = MKBADDR(myvolume);

	/* Now reply everything is ok */
	ReplyPacket(dp,DOSTRUE,DOSTRUE);

	/* Now forever or something like that */
	cont = TRUE;
	while (cont==TRUE) {
		/* Wait for incoming packets */
		WaitPort(mp);
		/* Get some packet */
		while ((mn=GetMsg(mp))!=NULL) {
			/* Get the DosPacket */
			dp = (struct DosPacket *)mn->mn_Node.ln_Name;
			/* What sort of action is required ? */
			switch (dp->dp_Type) {
			case ACTION_DISK_INFO:
				id = BADDR(dp->dp_Arg1);
				id->id_NumSoftErrors = 0;
				id->id_UnitNumber = 0;
				id->id_DiskState = ID_VALIDATED;
				id->id_NumBlocks = 100;
				id->id_NumBlocksUsed = 50;
				id->id_BytesPerBlock = 512;
				id->id_DiskType = ID_DOS_DISK;
				id->id_VolumeNode = (BPTR)MKBADDR(myvolume);
				id->id_InUse = DOSTRUE;
				ReplyPacket(dp,DOSTRUE,DOSTRUE);
				break;
			case ACTION_IS_FILESYSTEM:
				ReplyPacket(dp,DOSTRUE,DOSTRUE);
				break;
			case ACTION_LOCATE_OBJECT:
				fl = MyLock((BSTR)dp->dp_Arg2,
					(struct FileLock *)BADDR(dp->dp_Arg1),
					(LONG)dp->dp_Arg3,mp,myvolume);
				if (fl==NULL) {
					ReplyPacket(dp,DOSFALSE,
						ERROR_NO_FREE_STORE);
					break;
				}
				ReplyPacket(dp,(LONG)MKBADDR(fl),DOSTRUE);
				break;
			case ACTION_EXAMINE_OBJECT:
				fl = BADDR(dp->dp_Arg1);
				fib = BADDR(dp->dp_Arg2);
				if (fl->fl_Key==0) {
					fib->fib_DiskKey = fl->fl_Key+1;
					fib->fib_DirEntryType = ST_USERDIR;
					strcpy(fib->fib_FileName,"\3LHA");
					fib->fib_Protection = FIBB_READ;
					//fib->fib_Size = 1024;
					//fib->fib_NumBlocks = 2;
					// fib->fib_Date =
				}
				fib->fib_Comment[0] = '\0';
				fib->fib_EntryType = fib->fib_DirEntryType;
				ReplyPacket(dp,DOSTRUE,DOSTRUE);
				break;
			case ACTION_EXAMINE_NEXT:
				fl = BADDR(dp->dp_Arg1);
				fib = BADDR(dp->dp_Arg2);
				if (fib->fib_DiskKey<5) {
					fib->fib_DiskKey++;
					fib->fib_DirEntryType = ST_FILE;
					strcpy(fib->fib_FileName,"\7MyFile#");
					fib->fib_FileName[7] = '0'+fib->fib_DiskKey;
					fib->fib_Protection = FIBB_READ;
					fib->fib_Size = 1024;
					fib->fib_NumBlocks = 2;
					// fib->fib_Date =
					fib->fib_Comment[0] = '\0';
					fib->fib_EntryType = fib->fib_DirEntryType;
					ReplyPacket(dp,DOSTRUE,DOSTRUE);
				}
				else {
					ReplyPacket(dp,DOSFALSE,ERROR_NO_MORE_ENTRIES);
				}
				break;
			case ACTION_DIE:
				/* Destroy the volume entry */
				dl = BADDR(di->di_DevInfo);
				if (dl==myvolume) {
					di->di_DevInfo = myvolume->dl_Next;
				}
				else {
					while (dl!=NULL &&
						(struct DeviceList *)
						BADDR(dl->dl_Next)!=myvolume)
						dl = (struct DeviceList *)
							BADDR(dl->dl_Next);
					if (dl!=NULL) {
						dl->dl_Next = myvolume->dl_Next;
					}
				}
				FreeVec(myvolume);
				ReplyPacket(dp,DOSTRUE,DOSTRUE);
				/* I must disappear now!!! */
				dn->dn_Task = NULL;
				cont = FALSE;
				break;
			default:
				ReplyPacket(dp,DOSFALSE,ERROR_ACTION_NOT_KNOWN);
				break;
			}
		}
	}
	CloseLibrary((struct Library *)DosBase);
}

void	ReplyPacket(struct DosPacket *dp,LONG res1,LONG res2)
{
struct MsgPort	*mp;
struct Message	*mn;

	mp = dp->dp_Port;
	mn = dp->dp_Link;
	mn->mn_Node.ln_Name = (char *)dp;
	dp->dp_Port = &((struct Process *)FindTask(NULL))->pr_MsgPort;
	dp->dp_Res1 = res1;
	dp->dp_Res2 = res2;
	PutMsg(mp,mn);
}

char	*ToBStr(char *string)
{
char	*t;
int	l, i;

	for (l=0,t=string; *t!='\0'; l++,t++);
	t = AllocVec(l+1,MEMF_CLEAR|MEMF_CHIP);
	for (i=0; i<l; i++) t[i+1] = string[i];
	t[0] = (char)l;
	return t;
}

struct FileLock	*MyLock(BSTR name,struct FileLock *fl,LONG access,
		struct MsgPort *mp,struct DeviceList *dl)
{
struct FileLock	*newfl;
static int	key=0;

	newfl = (struct FileLock *)AllocVec(sizeof(struct FileLock),
		MEMF_CHIP|MEMF_CLEAR);
	if (newfl==NULL) return NULL;
	newfl->fl_Link = NULL;
	newfl->fl_Key = key++;
	newfl->fl_Access = access;
	newfl->fl_Task = mp;
	newfl->fl_Volume = MKBADDR(dl);
	return newfl;
}
