#include <exec/types.h>
#include <exec/devices.h>
#include <exec/errors.h>
#include <exec/execbase.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/filehandler.h>
#include <proto/exec.h>
#include <proto/dos.h>

#define __USE_SYSBASE

void	ReplyPacket(struct DosPacket *,LONG,LONG);
char	*ToBStr(char *);

extern struct ExecBase *__far	AbsExecBase;
struct ExecBase *SysBase;
struct DosLibrary	*DosBase;

void __saveds start(void)
{
struct MsgPort		*mp = NULL;
struct DosPacket	*dp = NULL;
struct DeviceNode       *dn = NULL;
struct DeviceList	*dl = NULL;
struct DosInfo		*di = NULL;
struct Message		*mn = NULL;
struct InfoData		*id = NULL;
struct FileInfoBlock	*fib = NULL;
struct FileLock		*fl = NULL;
int			cont;

        SysBase = AbsExecBase;
	DosBase = (struct DosLibrary *)OpenLibrary(DOSNAME,LIBRARY_MINIMUM);

	/* Get my Message Port */
	mp = &((struct Process *)FindTask(NULL))->pr_MsgPort;

	/* Wait for the Start-up Packet */
	WaitPort(mp);

	/* Now get the Start-up Packet */
	dp = (struct DosPacket *)GetMsg(mp)->mn_Node.ln_Name;

	/* dp->dp_Arg1 : BPTR to BSTR (device name) */
	/* dp->dp_Arg2 : dn->dn_Startup */
	/* dp->dp_Arg3 : BPTR to struct DeviceNode */
	/* dp->dp_Arg4 : struct MsgPort * (alternate handler port) */
        
	/* Get the DeviceNode */
        dn = BADDR(dp->dp_Arg3);
        dn->dn_Task = mp;

	/* Create the VolumeNode */
	dl = (struct DeviceList *)MakeDosEntry("JBY",DLT_VOLUME);
	if (dl!=NULL) {
		di = (struct DosInfo *)BADDR(DosBase->dl_Root->rn_Info);
		dl->dl_Task = mp;
		/*dl->dl_VolumeDate */
		dl->dl_LockList = NULL;
		dl->dl_DiskType = ID_DOS_DISK;
/*
		Forbid();
		dl->dl_Next = di->di_DevInfo;
		di->di_DevInfo = (BPTR)MKBADDR(dl);
		Permit();
*/
		LockDosList(LDF_VOLUMES|LDF_WRITE);
		AddDosEntry((struct DosList *)dl);
		UnLockDosList(LDF_VOLUMES|LDF_WRITE);
		Permit();
	}

	/* Now reply everything is ok */
	ReplyPacket(dp,DOSTRUE,DOSTRUE);

	/* Now forever or something like that */
	cont = TRUE;
	while (cont==TRUE) {
		/* Wait for incoming packets */
		WaitPort(mp);
		/* Get some packet */
		while ((mn=GetMsg(mp))!=NULL) {
			/* Get the DosPacket */
			dp = (struct DosPacket *)mn->mn_Node.ln_Name;
			/* What sort of action is required ? */
			switch (dp->dp_Type) {
			case ACTION_DISK_INFO:
				id = BADDR(dp->dp_Arg1);
				id->id_NumSoftErrors = 0;
				id->id_UnitNumber = 0;
				id->id_DiskState = ID_VALIDATED;
				id->id_NumBlocks = 100;
				id->id_NumBlocksUsed = 50;
				id->id_BytesPerBlock = 512;
				id->id_DiskType = ID_DOS_DISK;
				id->id_VolumeNode = (BPTR)MKBADDR(dl);
				id->id_InUse = DOSTRUE;
				ReplyPacket(dp,DOSTRUE,DOSTRUE);
				break;
			case ACTION_IS_FILESYSTEM:
				ReplyPacket(dp,DOSTRUE,DOSTRUE);
				break;
			case ACTION_LOCATE_OBJECT:
				fl = BADDR(dp->dp_Arg1);
				fl->fl_Link = NULL;
				fl->fl_Key = 0;
				fl->fl_Access = SHARED_LOCK;
				fl ->fl_Task = mp;
				fl ->fl_Volume = (BPTR)MKBADDR(dl);
				ReplyPacket(dp,(LONG)fl,DOSTRUE);
				break;
			case ACTION_EXAMINE_OBJECT:
				fib = BADDR(dp->dp_Arg2);
				fib->fib_DiskKey = 0;
				fib->fib_DirEntryType = ST_FILE;
				fib->fib_FileName[0] = 'C';
				fib->fib_FileName[1] = 'o';
				fib->fib_FileName[2] = 'u';
				fib->fib_FileName[3] = 'c';
				fib->fib_FileName[4] = 'o';
				fib->fib_FileName[5] = 'u';
				fib->fib_FileName[6] = '\0';
				fib->fib_Protection = NULL;
				fib->fib_EntryType = fib->fib_DirEntryType;
				fib->fib_Size = 1024;
				fib->fib_NumBlocks = 2;
				/* fib->fib_Date */
				fib->fib_Comment[0] = '\0';
				ReplyPacket(dp,DOSTRUE,DOSTRUE);
				break;
			case ACTION_EXAMINE_NEXT:
				ReplyPacket(dp,DOSFALSE,ERROR_NO_MORE_ENTRIES);
				break;
			case ACTION_DIE:
				ReplyPacket(dp,DOSTRUE,DOSTRUE);
				dn->dn_Task = NULL;
				cont = FALSE;
				break;
			default:
				ReplyPacket(dp,DOSFALSE,ERROR_ACTION_NOT_KNOWN);
				break;
			}
		}
	}
	if (dl!=NULL) {
		LockDosList(LDF_VOLUMES|LDF_DELETE|LDF_WRITE);
		RemDosEntry((struct DosList *)dl);
		UnLockDosList(LDF_VOLUMES|LDF_DELETE|LDF_WRITE);
		Permit();
		FreeDosEntry((struct DosList *)dl);
	}
	if (DosBase!=NULL) CloseLibrary((struct Library *)DosBase);
}

void	ReplyPacket(struct DosPacket *dp,LONG res1,LONG res2)
{
struct MsgPort	*mp;
struct Message	*mn;

	mp = dp->dp_Port;
	mn = dp->dp_Link;
	mn->mn_Node.ln_Name = (char *)dp;
	dp->dp_Port = &((struct Process *)FindTask(NULL))->pr_MsgPort;
	dp->dp_Res1 = res1;
	dp->dp_Res2 = res2;
	PutMsg(mp,mn);
}

char	*ToBStr(char *string)
{
char	*t;
int	l, i;

	for (l=0,t=string; *t!='\0'; l++,t++);
	t = AllocVec(l+1,MEMF_CLEAR|MEMF_CHIP);
	for (i=0; i<l; i++) t[i+1] = string[i];
	t[0] = (char)l;
	return t;
}
